/* =============================================================

	File: msg.c

	Title: Message Handler

	Author: Ashok

	Purpose: Displays messages

================================================================ */

	#include <stdio.h>
	#include <stdlib.h>
	#include <stdarg.h>
	#include "msg.h"
	#include "xpbasic.h"

#if !defined(NDEBUG)
	static FILE *msgfile=stdout, *errfile=stderr, *logfile=stdout;
#endif

void msg(char *fmt, ...)
{
#if !defined(NDEBUG)
	va_list args;

	va_start(args, fmt);
	fprintf(msgfile, "Message: ");
	vfprintf(msgfile, fmt, args);
	va_end(args);
	fputc('\n', msgfile);
#endif
}

void warn(char *fmt, ...)
{
#if !defined(NDEBUG)
	va_list args;

	va_start(args, fmt);
	fprintf(errfile, "WARNING: ");
	vfprintf(errfile, fmt, args);
	va_end(args);
	fputc('\n', errfile);
#endif
}

void pwarn(char *fmt, ...)
{
#if !defined(NDEBUG)
	va_list args;

	va_start(args, fmt);
	fprintf(errfile, "WARNING: ");
	vfprintf(errfile, fmt, args);
	va_end(args);
	perror("\n");
	fputc('\n', errfile);
#endif
}

void err(char *fmt, ...)
{
#if !defined(NDEBUG)
	va_list args;

	va_start(args, fmt);
	fprintf(errfile, "FATAL ERROR: ");
	vfprintf(errfile, fmt, args);
	va_end(args);
	fputc('\n', errfile);
	exit(1);
#endif
}

void perr(char *fmt, ...)
{
#if !defined(NDEBUG)
	va_list args;

	va_start(args, fmt);
	fprintf(errfile, "FATAL ERROR: ");
	vfprintf(errfile, fmt, args);
	va_end(args);
	perror("\n");
	fputc('\n', errfile);
	exit(1);
#endif
}

void log_open(char *filename)
{
#if !defined(NDEBUG)
	logfile = fopen(filename, "at");
	if (logfile==NULL)
	{
		msg("Unable to open log file %s\n", filename);
		logfile = stdout;
	}
	atexit(log_close);
	log("Log opened at %s\n", ctime(time(0)));
#endif
}

void log(char *fmt, ...)
{
#if !defined(NDEBUG)
	va_list args;

	va_start(args, fmt);
	vfprintf(logfile, fmt, args);
	va_end(args);
	fputc('\n', logfile);
	fflush(logfile);
#endif
}

void log_close(void)
{
#if !defined(NDEBUG)
	if (logfile != stdout)
	{
		log("Log closed at %s\n", ctime(time(0)));
		close(logfile);
		logfile=stdout;
	}
#endif
}

/* Log info about a packet */
void logpkt(pkt *p, char *s)
{
#ifndef NDEBUG
	log("%s: size=%3d %8p-<%8p>-%8p ts=%8u-%8u gop_num=%4d type=%c fr_num=%d fr_part=%d_of_%d frag=%d_of_%d",  
						s, p->size, 
						p->prev, p, p->next,
						(unsigned )p->tstampl, (unsigned )p->tstamph, 
						p->gop_num, 
						p->curr_type,
						p->frame_num, 
						p->frame_part, 
						p->num_of_parts,
						p->frag_num, 
						p->total_frags);
#endif
}
