/* ======================================================

File: pktpool.c

Title: Packet Pool 

Author: Ashok

Purpose: 
Used for providing free packets to the transport. This module 
will maintain a pool of free packets to reduce the number of
calls to malloc and free. 

========================================================== */

#include "xpbasic.h"
#include "pktpool.h"

void setup_pool(void)
{
	int i; pkt *tmp;

	/* Put a header packet into each pool */
	free_first=free_last=(pkt *)malloc(sizeof(pkt));
	free_first->pool_prev=free_first->pool_next=0;
	used_first=used_last=(pkt *)malloc(sizeof(pkt));
	used_first->pool_prev=used_first->pool_next=0;

	/* Add packets into the free pool if necessary */
	if (INIT_POOL>0)
	{
		for (i=0; i<INIT_POOL; i++)
		{
			tmp=(pkt *)malloc(sizeof(pkt));
			free_last->pool_next=tmp;
			tmp->pool_prev=free_last;
			tmp->pool_next=tmp->prev=tmp->next=0;
			free_last=tmp;
		}
	}
}

void shutdown_pool(void)
{
	pkt *tmp;

	while (used_first)
	{
		tmp=used_first;
		used_first=used_first->pool_next;
		free(tmp);
	}

	while (free_first)
	{
		tmp=free_first;
		free_first=free_first->pool_next;
		free(tmp);
	}

	used_first=used_last=0;
	free_first=free_last=0;

}

pkt *pktalloc(void)
{
	static pkt *tmp;

	if (free_first == free_last)
	{
		/* Add more packets into the free pool */
		int i; 

		for (i=0; i<INC_POOL; i++)
		{
			tmp=(pkt *)malloc(sizeof(pkt));
			free_last->pool_next=tmp;
			tmp->pool_prev=free_last;
			tmp->pool_next=tmp->prev=tmp->next=0;
			free_last=tmp;
		}
	}

	/* Pull a packet off the top of the free list */
	tmp=free_first;
	free_first=tmp->pool_next;
	free_first->prev=0;

	/* Now add it to the bottom of the used list */
	tmp->pool_prev=used_last;
	used_last->pool_next=tmp;
	tmp->pool_next=0;
	used_last=tmp;
	return tmp;
}

void pktfree(pkt *p)
{
	/* Take the packet out of the used list */
	p->pool_prev->pool_next=p->pool_next;
	if (p->pool_next) p->pool_next->pool_prev=p->pool_prev;

	/* ... and put it in the free list */
	free_last->pool_next=p;
	p->pool_prev=free_last;
	p->pool_next=0;
	free_last=p;
}

void print_pool(void)
{
	pkt *tmp;

	printf("\nUsed list\n");
	for (tmp=used_first; tmp; tmp=tmp->pool_next)
		printf("%p ", tmp);

	printf("\nFree list\n");
	for (tmp=free_first; tmp; tmp=tmp->pool_next)
		printf("%p ", tmp);
}
