/* ======================================================

File: pktpool.h

Title: Packet Pool 

Author: Ashok

Purpose: 
Used for providing free packets to the transport. This module 
will maintain a pool of free packets to reduce the number of
calls to malloc and free. 

========================================================== */

#if !defined(_PKTPOOL)
#define _PKTPOOL

#include "xpbasic.h"

/* Number of packets to initially put in the free pool */
#define INIT_POOL 5

/* Number of packets to add to free pool, if it runs out */
#define INC_POOL 5

pkt *used_first, *used_last, *free_first, *free_last;

void setup_pool(void);
/* Sets up the packet pool */

void shutdown_pool(void);
/* Shuts down the packet pool and frees all memory */

pkt *pktalloc(void);
/* Allocates a packet from the pool */

void pktfree(pkt *p);
/* Returns a packet to the free pool */

#endif
