/* ===========================================================

File: player2.c

Title: Video Player v2

Author: Ashok

Purpose:
This module plays out packets from the top of the queue

=============================================================== */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/wait.h>
#include <assert.h>

#include "msg.h"
#include "xpbasic.h"
#include "pktpool.h"
#include "vrec2.h"

#define PLAYER "/usr/local/bin/mtvp"
#define PLAYER_ARGS "-"

/*
#define MTVBUFNAME "MTVP_SYS_BUFFER_SIZE"
#define MTVBUFSIZE "80000"
*/

/*
#define PLAYER "mpeg_play" 
#define PLAYER_ARGS "-framerate", "30", "-controls", "on", "stdin" 
*/
int playout_enable;
int pipefd;

static int player_pid;
static int fd[2] = {-1, -1};
static int packet_waiting = 0;
static pkt *packet;
static int flag_stop = 0;
static int last_gop_num = 0;
static int starved = 0;

void write_to_file(void)
{
	fd[1] = open("out.mps", O_CREAT | O_TRUNC | O_WRONLY);
	if (fd[1] <0)
		err("Unable to open file\n");
}

void start_player(void)
{
	if (pipe(fd))
		perr("Cannot get a pipe");

	if ((player_pid = fork()) < 0)
		perr("Unable to fork");

	if (player_pid==0)
	{
		/* client */
		
		/* First close the write end of the pipe */
		close(fd[1]);
		
		/* Now, if fd[0] is not stdin, make it so */
		if (fd[0] != STDIN_FILENO)
		{
			if (dup2(fd[0], STDIN_FILENO) != STDIN_FILENO)
				perr("dup2 in the client failed");
			close(fd[0]);
		}

#ifdef MTVBUFNAME
		setenv(MTVBUFNAME, MTVBUFSIZE, 1);
#endif
		
		/* ready to go! start up the player */
		if (execl(PLAYER, PLAYER, PLAYER_ARGS, 0)<0)
			perr("Unable to EXEC player %s", PLAYER);
	}
	else
	{
		/* server */

		/* Here we need to do nothing except close the read descriptor */
		close(fd[0]);
		fd[0] = -1;

		pipefd = fd[1];
		
		/* Make the write descriptor nonblocking */
		/*
		if (fcntl(fd[1], F_SETFL, O_NDELAY) < 0)
			perr("Unable to make the pipe nonblocking");
			*/
	}
}

void stop_player(void)
{
	if (fd[1]>=0)
	{
		flag_stop = 1;
	}
}

void pkt_dump(pkt *p)
{
	do 
	{
		assert(p!=NULL);
		logpkt(p, "GOP Dump");
		p=p->next;
	} while (p!=NULL);
	log("-----------------------------------------------------------------------------------------------");
}

void player_do(void)
{
	static int nbytes, nwritten, i;
	pkt *tm1;

	/*	for (i=0; i<5; i++) */
	{
		if (!packet_waiting)
		{
			/* Starving */
			if (curr_gop_hdr == first->next)
			{
				if (!starved)
				{
#ifndef NDEBUG
					putchar('*'); fflush(stdout);
#endif
					starved = 1;
				}
				return;
			}
			else starved = 0;
			
			while ( 0 && second_gop_hdr == first->next)
			{
				tm1 = second_gop_hdr;
				parse_gop();
				n_gop--; 
#if 0
				printf("<%d> ", n_gop); fflush(stdout);
#endif
				if (tm1==second_gop_hdr)
				{
					pkt_dump(second_gop_hdr);
					exit(1);
				}
				if (curr_gop_hdr == first->next)
				{
					/* Starving */
					return;
				}
			} /* while */
			packet = first->next;
#ifndef NDEBUG
			if (packet->gop_num != last_gop_num)
			{
#if 0
				printf("-%d ", --n_gop); fflush(stdout);
#endif
				last_gop_num = packet->gop_num;
			}
#endif
			/* logpkt(packet, "Set up"); */
		} /* if (!packet_waiting) */

		if (packet==NULL) return;

		nbytes = packet->size;
		/* logpkt(packet, "Playing"); */
		if (nbytes!=0)
		{
			nwritten = write(fd[1], (char *)packet->out_data, nbytes);
			if (nwritten <= 0 && errno == EAGAIN) 
			{
				packet_waiting = 1;
				return;
			}

			if (nwritten != nbytes)
				perr("Data not completely taken by player socket");
		}

		assert(packet->prev != NULL);
		assert(packet->next != NULL);
		packet->prev->next = packet->next;
		if (packet->next) packet->next->prev = packet->prev;
		pktfree(packet);
		packet_waiting = 0;
		if (first->next==NULL) last = first;
	} /* for ... */
}

void player_setup(void)
{
	start_player();
	/* write_to_file(); */
	packet_waiting = 0;
}

void player_shutdown(void)
{
	close(fd[1]);
	fd[1] = -1;
}
