#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <errno.h>
#include <assert.h>
#include <netdb.h>
#include <msg.h>
#include <sys/time.h>

#define PORT 12345
#define INT_GOPS 50
#define MAX_GOPS 1100
#define MPEG_FILE "file.mpg"
#define MC_ADDR "227.0.0.1"

void main(int argc, char *argv[])
{
	FILE *fd = popen("tee recvstat.out", "wt");
	FILE *mpg = fopen(MPEG_FILE, "wb"); 
	struct timeval start_time, end_time;
	struct sockaddr_in addr;
	unsigned char buf[576]; int buflen=576;
	struct ip_mreq mreq;
	int addrlen, nbytes;
	int sockd; int i;
	float t;
	int receiving = 0, latched=0, gop_num, frame_num, last_gop_num, total_size;

	if ((sockd=socket(AF_INET, SOCK_DGRAM, 0))<0)
		perr("Unable to create socket");

	memset(&addr, 0, sizeof(addr));
	addr.sin_family=AF_INET;
	addr.sin_addr.s_addr=htonl(INADDR_ANY); 
	addr.sin_port=htons(PORT);
	addrlen=sizeof(addr);
    
  /* Bind the socket to the address */
	if (bind(sockd,(struct sockaddr *) &addr,sizeof(addr)) < 0) 
		pwarn("I am unable to bind socket to address.");

	mreq.imr_multiaddr.s_addr=inet_addr(MC_ADDR);
	mreq.imr_interface.s_addr=htonl(INADDR_ANY);
	if (setsockopt(sockd,IPPROTO_IP,IP_ADD_MEMBERSHIP,&mreq,sizeof(mreq)) < 0) 
		pwarn("Unable to join multicast group.");

	while (1)
	{
		nbytes = recvfrom(sockd, buf, buflen, 0, (struct sockaddr *)&addr, &addrlen);
		if (nbytes<1)
			perr("Error in socket");
		if (nbytes<18)
			warn("Illegal packet");

		fwrite(buf+18, nbytes-18, 1, mpg);

		if (!receiving)
		{
			log("Packet received. Timer begun");
			gettimeofday(&start_time, 0);
			gop_num=0; last_gop_num=0; frame_num=0; total_size=0;
			receiving = 1;
		}

		total_size+=nbytes;
		
		if (buf[5]==1 && buf[7]==1)
		{
			/* New frame */
			if ((i=((int)buf[2] << 8 ) + buf[3]) != last_gop_num)
			{
				/* New GOP */
				last_gop_num=i;
				gop_num++;
				if (!(gop_num % INT_GOPS))
				{
					/* Print results */
					gettimeofday(&end_time, 0);
					i=end_time.tv_sec * 1000000 + end_time.tv_usec;
					i -= (start_time.tv_sec * 1000000 + start_time.tv_usec);
					log("\nINTERMEDIATE RESULTS: %d gops received\n", gop_num);
					fprintf(fd, "Number of GOPS received: %d\n", gop_num);
					fprintf(fd, "Number of (non-A) frames received: %d\n", frame_num);
					fprintf(fd, "Number of bytes received: %d\n", total_size);
					t=(float)i/1000000;
					fprintf(fd, "Total time taken: %10.3f sec\n", t);
					fprintf(fd, "\nTransfer rate (bits/sec): %f\n", (float)(total_size*8)/t);
					fprintf(fd, "Transfer rate (bytes/sec): %f\n", (float)total_size/t);
					fprintf(fd, "Transfer rate (non-A frames/sec): %f\n", (float)frame_num/t);
					fprintf(fd, "Transfer rate (gops/sec): %f\n", (float)gop_num/t);
				}
				if (gop_num>MAX_GOPS)
					break;
				fprintf(fd, "\n Gop %d: ", last_gop_num); fflush(stdout);
			}
			if (buf[9] != 'A')
			{
			 	frame_num++;
				fputc(buf[9], fd); fflush(fd);
			}
		}

	}

	/* Print results */
	gettimeofday(&end_time, 0);
	i=end_time.tv_sec * 1000000 + end_time.tv_usec;
	i -= (start_time.tv_sec * 1000000 + start_time.tv_usec);
	log("\n\nFINAL RESULTS: %d gops received\n", gop_num);
	fprintf(fd, "Number of GOPS received: %d\n", gop_num);
	fprintf(fd, "Number of (non-A) frames received: %d\n", frame_num);
	fprintf(fd, "Number of bytes received: %d\n", total_size);
	t=(float)i/1000000;
	fprintf(fd, "Total time taken: %10.3f sec\n", t);
	fprintf(fd, "\nTransfer rate (bits/sec): %f\n", (float)(total_size*8)/t);
	fprintf(fd, "Transfer rate (bytes/sec): %f\n", (float)total_size/t);
	fprintf(fd, "Transfer rate (non-A frames/sec): %f\n", (float)frame_num/t);
	fprintf(fd, "Transfer rate (gops/sec): %f\n", (float)gop_num/t);

	exit(0);
}


	
	
