/* ===========================================================

File: vrec2.c

Title: Video Receiver v2

Author: Ashok

Purpose:
This module receives and resequences the video packets
from the network. 

=============================================================== */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <errno.h>
#include <assert.h>
#include <netdb.h>

#include "msg.h"
#include "xpbasic.h"
#include "pktpool.h"
#include "player2.h"
#include "client.h"
#include "vrec2.h"

pkt *first, *last;
pkt *curr_gop_hdr, *second_gop_hdr;
int n_gop=0;

static unsigned long vid_serv_addr;

static struct sockaddr_in addr;
int vsock_fd = 0;
static int addrlen;
static struct ip_mreq mreq;
static int latched;

void printq(void);
void close_video_socket(void);

void decode_packet(pkt *p)
{
	int i=0;
	unsigned char *buf = p->data;

	p->out_data = p->data + 18;
	p->size = (short)buf[i++]; p->size <<= 8;
	p->size += (short)buf[i]; i++;
	p->gop_num = (short)buf[i++]; p->gop_num <<= 8;
	p->gop_num += (short)buf[i]; i++;
	p->frame_num = buf[i++];
	p->frame_part = buf[i++];
	p->num_of_parts = buf[i++];
	p->frag_num = buf[i++];
	p->total_frags = buf[i++];
	p->curr_type = buf[i++];
	p->tstamph   = (short)buf[i]; p->tstamph <<= 8; i++;
	p->tstamph  += (short)buf[i]; p->tstamph <<= 8; i++;
	p->tstamph  += (short)buf[i]; p->tstamph <<= 8; i++;
	p->tstamph  += (short)buf[i]; i++;
	p->tstampl   = (short)buf[i]; p->tstampl <<= 8; i++;
	p->tstampl  += (short)buf[i]; p->tstampl <<= 8; i++;
	p->tstampl  += (short)buf[i]; p->tstampl <<= 8; i++;
	p->tstampl  += (short)buf[i];
} 

void flush_queues(void)
{
	pkt *p1=first->next, *p2; int i;

	/* if (p1==NULL) */
	{
		n_gop=playout_enable=0;
		while (p1!=NULL)
		{
			p2 =p1;
			p1 = p1->next;
			pktfree(p2);
		}
		last=first;
		curr_gop_hdr=second_gop_hdr=0;
		latched=0;
		return;
	}

	for (i=0; i<N_GOP; i++) 
	{
		while (!changedGOP(p1))
		{
			if (p1==curr_gop_hdr)
				break;
			assert(p1->next!=NULL);
			p1=p1->next;
		}
		if (p1==curr_gop_hdr) break;
		p1=p1->next;
	}

	p2=p1->next;
	p1->next=NULL; last=p1;

	while (p2!=NULL)
	{
		pktfree(p2);
		p2=p2->next;
	}
	n_gop=i;
	curr_gop_hdr = 0;
	latched=0;
	second_gop_hdr = first->next;
}

void init_video_socket()
{
	if (!first)
	{
		first=last=pktalloc();
		first->tstampl = first->tstamph =0;
	}

	/* Create a UDP socket */
	if ((vsock_fd=socket(AF_INET, SOCK_DGRAM, 0))<0)
		perr("Unable to create socket for video.");

	log("UDP Socket created: %d", vsock_fd);
	/* Set the address */
	memset(&addr, 0, sizeof(addr));
	addr.sin_family=AF_INET;
	addr.sin_addr.s_addr=htonl(INADDR_ANY); 
	addr.sin_port=htons(VIDEO_PORT);
	addrlen=sizeof(addr);
    
  /* Bind the socket to the address */
	if (bind(vsock_fd,(struct sockaddr *) &addr,sizeof(addr)) < 0) 
		pwarn("I am unable to bind socket to address.");

	/* Make the socket nonblocking */
	/*	
	if (fcntl(vsock_fd, F_SETFL, O_NONBLOCK)<0)
		pwarn("Unable to make the socket nonblocking");
		*/

	log("UDP Socket created");
}

void close_video_socket(void)
{
	close(vsock_fd);
	log("UDP Socket %d closed", vsock_fd);
	vsock_fd = 0;
}

void vrec_join(unsigned long srv_multicast_addr)
{
	/* Leave current group */
	if (vid_serv_addr != 0)
		vrec_leave();
	
	if (srv_multicast_addr == 0)
	{
	    log("Turning off reception");
		vid_serv_addr=0;
		return;
		
	}

	/* Request to join a multicast group */
	mreq.imr_multiaddr.s_addr=srv_multicast_addr;
	mreq.imr_interface.s_addr=htonl(INADDR_ANY);
	if (setsockopt(vsock_fd,IPPROTO_IP,IP_ADD_MEMBERSHIP,&mreq,sizeof(mreq)) < 0) 
		pwarn("Unable to join multicast group.");

	vid_serv_addr = srv_multicast_addr;

	log("Joined multicast group ");
}

void vrec_leave()
{
	mreq.imr_multiaddr.s_addr=vid_serv_addr;
	mreq.imr_interface.s_addr=htonl(INADDR_ANY);
	if (setsockopt(vsock_fd,IPPROTO_IP,IP_DROP_MEMBERSHIP, &mreq, sizeof(mreq)) < 0) 
		pwarn("Unable to leave multicast group.");
	log("Left current group");
	vid_serv_addr = 0;
	flush_queues();
}

/* Checks if a is the first packet in a new GOP */
int changedGOP(pkt *a)
{
	if (a->prev==first) return 1;
	return a->prev->gop_num!=a->gop_num;
}

/* Dumps all packets between p1 and p2 (both inclusive)
with no regard to type */
void dump_pkts_all(pkt *p1, pkt *p2)
{
	assert(p1->prev != NULL);
	assert(p1!=NULL);
	assert(p2!=NULL);

	while (p1!=p2->next)
	{
		assert(p1!=NULL);
		p1->prev->next = p1->next;
		if (p1->next!=NULL)
			p1->next->prev = p1->prev;
		pktfree(p1);
		p1=p1->next;
	}
}

/* Dumps all packets between p1 and p2 of type t */
void dump_pkts(pkt *p1, pkt *p2, unsigned char t)
{
	assert(p1->prev != NULL);
	assert(p1!=NULL);
	assert(p2!=NULL);

	while (p1!=p2->next)
	{
		assert(p1 != NULL);
		if (p1->curr_type==t)
		{
			p1->prev->next=p1->next;
			if (p1->next!=NULL)
				p1->next->prev = p1->prev;
			pktfree(p1);
		}
		p1=p1->next;
	}
}

/* This function verifies an entity starting from
pin. The entity type, frame number and frame part are
given - all fragments in this entity should
match this info. A pointer to the starting fragment of
the next entity is returned in pout. The function
returns 1 if the entity is validated, 0 otherwise.
This function makes no changes to the queue */
int verify_entity(pkt *pin, unsigned int mygopnum, unsigned char mytype, int myfnum, int myfpart, pkt **pout)
{
	int i, j;

	assert(pin != NULL);
	assert(pin->prev != NULL);

	*pout = pin;
	j = pin->total_frags;
	for (i=1; i<=j; i++)
	{
		if (  (*pout)->frag_num != i ||
					(*pout)->gop_num != mygopnum ||
					(*pout)->curr_type != mytype ||
					(*pout)->frame_num != myfnum ||
					(*pout)->frame_part != myfpart )
		{
			/* Bad entity - find the end */
			while ((*pout)->gop_num == mygopnum &&
						 (*pout)->curr_type == mytype &&
						 (*pout)->frame_num == myfnum &&
						 (*pout)->frame_part == myfpart )
				*pout = (*pout)->next;
			return 0;
		}
		*pout = (*pout)->next;
 	}
	return 1;
}

/* This function parses & fixes the GOP starting at
first->next. It repositions the second_gop_hdr to
the start of the next gop. Any packets which can't be
given to the player are deleted - this might be
the whole gop if necessary. 

Rules for gop fixing:

1) If ANY frame is incomplete, the ENTIRE frame
must be dropped.

2) If a P frame is incomplete, all B frames
imeediately to the left of  it are dropped.

3) If an I frame is incomplete, the entire gop
goes.  */

void parse_gop(void)
{
	pkt *p1, *p2, *p3, *p4, *p5, *p6;
	int i, j;

	assert(first->next != curr_gop_hdr); /** Can't parse incomplete GOP */
	assert(first->next != NULL); /* Obviously */
	assert(first->next->next != NULL); /* Obviously */

	p1 = first->next;
	if (p1->gop_num==0)
		p1=p1->next;
	do
	{
		for (p2=p1, i=1; i <= p1->num_of_parts; i++)
		{
			if (p2->curr_type != (p2->prev)->curr_type )
			{
				if (p2->curr_type!='A') 
					p4=p2;
				else
				{
					p6=p2;
					i--;
				}
			}

			if (!verify_entity(p2, p2->gop_num, p2->curr_type, p2->frame_num, p2->frame_part, &p3))
			{
				switch(p2->curr_type)
				{
					case 'A': /* drop this A frame */
						for (p5=p3; !changedGOP(p5) && p5->curr_type==p2->curr_type && p5->frame_num==p2->frame_num; p5=p5->next)
							assert(p5!=NULL);
						log("A frame bad - dropping packets %p to %p", p6, p5->prev);
						dump_pkts(p6, p5->prev, 'A');
						p2=p5;
						break;

					case 'B': /* drop this B frame, preserving any A frames in the way */
						for (p5=p3; !changedGOP(p5) && p5->curr_type==p2->curr_type && p5->frame_num==p2->frame_num; p5=p5->next)
							assert(p5!=NULL);
						log("B frame bad - dropping packets %p to %p", p4, p5->prev);
						dump_pkts(p4, p5->prev, 'B');
						p2=p5;
						break;

					case 'I': /* drop everything */
						for (p2=first->next->next; !changedGOP(p2); p2=p2->next)
							assert(p2!=NULL); /* broken/incomplete queue */
						second_gop_hdr = p2;
						log("I frame bad - dropping packets %p to %p", first->next, p2->prev);
						dump_pkts_all(first->next, p2->prev);
						return;

					case 'P': /* drop this P, any previous B, and everything after this */
						p2=p1->next;
						while (!changedGOP(p1) && (p1->prev->curr_type == 'B' || p1->prev->curr_type == 'A'))
						{
							p1=p1->prev;
							assert(p1->prev!=NULL); 
							if (p1==first->next) break;
						}
						while (!changedGOP(p2))
						{
							p2=p2->next;
							assert(p2!=NULL);
						}

						log("P frame bad - dropping packets %p to %p", p1, p2->prev);
						dump_pkts(p1, p4->prev, 'B');
						dump_pkts_all(p4, p2->prev);
						second_gop_hdr = p2;
						return;

					default: /* Unknown packet type */
						warn("Unexpected packet type %c\n", p2->curr_type);
						p2=p3;
				} /* switch */
			} /* if (!verify_... */
			else p2=p3;
		} /* for (p2 = ... */
		p1=p2;
	} while (!changedGOP(p1));

	second_gop_hdr = p1;
	return;
} /* parse_gop() */	

void printq(void)
{
	pkt *p=first;
	while (p!=0)
	{
		logpkt(p,"Queue   ");
		p=p->next;
	}
}

/* Returns the system header & video sequence header for player resync */
void get_headers(char *dat, int *len)
{
	memcpy(dat, SYSHDR, SYSHDR_LEN);
	*len = SYSHDR_LEN;
}

void get_video_packet(void)
{
	static int nbytes=0;
	static pkt *tmp=0, *tmp2=0;
	static int kk;

	if (!tmp) tmp=pktalloc();
	/* Attempt to read from the socket */
	nbytes = recvfrom(vsock_fd, &(tmp->data), PKT_SIZE, 0, 0, 0);
	if (nbytes<1)
	{
		if (errno ==  EWOULDBLOCK)
			return;
		else perr("Error in recvfrom.");
	}

	decode_packet(tmp);

	if (!latched)
	{
		if ((tmp->curr_type=='I' &&
				tmp->frame_num==1 &&
				tmp->frame_part==1 &&
				tmp->frag_num==1) ||
				(tmp->gop_num==0))
		{
			log("Latched on to GOP start");
			latched=1;
 			tmp2=pktalloc(); 
 			get_headers(tmp2->data, &(tmp2->size));
 			tmp2->out_data = tmp2->data; 
			tmp2->tstamph = tmp2->tstampl = 0; 
			tmp2->gop_num=0;
			last->next=tmp2;
			tmp2->prev=last;
			tmp2->next=NULL; 
			last=tmp2; 
			curr_gop_hdr = tmp2; 
 			log("Written headers. Length=%d", tmp2->size); 
			
			if (tmp->gop_num==0)
				return;
		}
		else
			return;
	}

	logpkt(tmp, "Received");
	/* Find this packet's position in the queue */

	tmp2=last;
	while (tmp2!=first && 
				(	(tmp2->tstamph > tmp->tstamph) || 
					(tmp2->tstamph == tmp->tstamph && tmp2->tstampl > tmp->tstampl)))
		 tmp2=tmp2->prev;
	tmp->next=tmp2->next;
	tmp->prev=tmp2;
	tmp2->next=tmp;
	if (tmp->next) tmp->next->prev=tmp;
	if (tmp2==last) last=tmp;

	if (curr_gop_hdr==0)
		curr_gop_hdr=tmp;

	/* Is it a GOP header? If so, restart the packet counter */
	if (changedGOP(tmp))
	{
		n_gop++; 
#ifndef NDEBUG
#if 0
		printf("%d ", n_gop); fflush(stdout); 
#endif
#endif
		curr_gop_hdr = tmp;
	}

	tmp=pktalloc();

	if (n_gop==N_GOP+1)
	{
		if (second_gop_hdr == 0) second_gop_hdr = first->next;
		playout_enable = 1;
	}
}

void vrec_do(void)
{
	get_video_packet();	
}

void vrec_setup()
{
	n_gop = 0;
	second_gop_hdr  = curr_gop_hdr = 0;
	latched=0;
}

void vrec_exit(void)
{
	close_video_socket();
}

