#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/errno.h>
#include <fcntl.h>
#include <signal.h>
#include <netdb.h>

#include "cchannel.h"
#include "dsap.h"
#include "agent.h"
#include "issue.h"
#include "path.h"
#include "vcripc.h"

#define MG_ADD_VALUE (1<<24)

FILE *fperr;
static int sockfd;
extern int errno;


void send_to_client(unsigned long hostid, char code, unsigned long mc_addr)
{
	char data[5];
	struct sockaddr_in addr;
	int sockfd;

	sockfd = socket(AF_INET, SOCK_DGRAM, 0);
	if (sockfd < 0)
		err_sys("Cannot create socket");
	
	data[0] = code;
	if (mc_addr>0)
		LONG2BYTES(mc_addr, data+1);
		
	bzero((struct sockaddr *)&addr, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = hostid;
	addr.sin_port = htons(DSAP_OUT_PORTNUM);
		
	if (sendto(sockfd, data, mc_addr?5:1, 0, (struct sockaddr *)&addr, sizeof(addr))<0)
		err_sys("Unable to send data to client");
/* 	else err_msg("Sent data to client"); */
}

void read_movies()
{
	char s[512];
	char nm[512];
	FILE *f = fopen("movies.dat", "r");
	int i;

	if (f==NULL)
		err_sys("Unable to open movie data file");
	num_movies = 0;

	while (fgets(s, 511, f)!=NULL)
	{
		if (s[0]=='\n')
	    continue;

		sscanf(s, " %s", nm);
		movielist[num_movies] = malloc(strlen(nm)+2);
		strcpy(movielist[num_movies], nm);
		num_movies++;
	}

	fclose(f);

	/*	for (i=0; i<num_movies; i++)
		printf("MOVIE: %s\n", movielist[i]); */
}

unsigned long generate_mcast_group()
{
	unsigned long num;
	stream_t *s1;
	int flag;

	for (num = inet_addr(BASE_MCAST_ADDR); ; num += MG_ADD_VALUE)
	{
		for (flag=0, s1 = stream_list; s1!=NULL; s1=s1->next)
	    if (s1->stream_id == num)
	    {
				flag =1;
				break;
	    }
		if (flag==0)
	    return num;
	}
}


unsigned long return_mcast_group(int movie_id) 
{
	stream_t *st = stream_list;

	while (st != NULL) 
	{
		if (st->movie_id == movie_id)
			return (st->stream_id);
     
		st = st->next;
	}
 
	return 0;
}


stream_t *create_stream(int movie_id, unsigned long start_frame_num)
{
	stream_t *s1, *s2;
		
	s1 = (stream_t *)malloc(sizeof(stream_t));
	if (s1==NULL)
		err_sys("Unable to allocate stream_t structure ");

	s1->stream_id = generate_mcast_group();
	s1->movie_id = movie_id;
	s1->frame_num = start_frame_num;
	s1->clients = NULL;
	s1->next = stream_list;
	stream_list = s1;
	return s1;
}

void disable_stream(unsigned long stream_id)
{
	stream_t *s1;

/* 	err_msg("Trying to disable stream %d", stream_id); */
	for (s1 = stream_list; s1!=NULL; s1=s1->next)
	{
		if (s1->stream_id == stream_id)
		{
/* 	    err_msg("Stream %d disabled", stream_id); */
	    mark_entry(stream_id, S_ACTIVE|S_PAUSED, S_STOPREQUEST);
	    s1->movie_id = -1;
	    return;
		}
	}
}


void remove_stream(unsigned long stream_id)
{
	stream_t *s1 = stream_list;

	if (stream_list==NULL)
		return;

	if (stream_list->stream_id == stream_id)
	{
		stream_list = stream_list -> next;
		free(s1);
	}
	else 
	{
		while ( s1->next != NULL)
		{
	    if (s1->next->stream_id == stream_id)
				break;
	    s1=s1->next;
		}
				
		if (s1->next!=NULL)
		{
	    stream_t *s2 = s1->next;
	    s1->next = s1->next->next;
	    free(s2);
		}
	}

	/*	remove_entry(stream_id);*/
}


void add_client_to_stream(stream_t *stream, unsigned long hostid)
{
	client_t *c = stream->clients;
	client_t *c1 = (client_t *)malloc(sizeof(client_t));
	if (c1==NULL)
		err_sys("Unable to allocate a client_t object");

	c1->hostid = hostid;
	c1->next = NULL;

	if (c==NULL)
	{
/* 		err_msg("First client for stream"); */
		stream->clients = c1;
	}
	else
	{
		c1->next = stream->clients;
		stream->clients = c1;
	}

/* 	err_msg("Client added. Sending message to client"); */
	send_to_client(hostid, CC_SWITCH_CHANNEL, stream->stream_id);
}			


void remove_client_from_stream(stream_t *stream, unsigned long hostid)
{
	client_t *c = stream->clients;
	client_t *c1;
		
	if (c==NULL) 
		return;

	if (c->hostid == hostid)
	{
		stream->clients = c->next;
		free(c);
		return; 
	}
		
	while (c->next!=NULL)
	{
		if (c->next->hostid == hostid)
	    break;
		c=c->next;
	}

	if (c->next!=NULL)
	{
		c1 = c->next;
		c->next = c->next->next;
		free(c1);
	}
}


stream_t *find_stream(unsigned long mcast_addr)
{
	stream_t *s;
	for (s=stream_list; s!=NULL; s=s->next)
	{
		if (s->stream_id == mcast_addr)
			return s;
	}

	return NULL;
}


stream_t *find_client(unsigned long hostid)
{
	stream_t *s;
	client_t *c;

	for (s=stream_list; s!=NULL; s=s->next)
	{
		for (c=s->clients; c!=NULL; c=c->next)
		{
	    if (c->hostid==hostid)
				return s;
		}
	}

	return NULL;
}



void remove_all_clients_from_stream(unsigned long stream_id)
{
	client_t *c, *p1, *p2, *tmp;
	stream_t *st;
	
	st = find_stream(stream_id);
	c = st->clients;

	p1 = c;
	p2 = c?c->next:NULL;

	while (p1)
	{
		if (p1 && !p2)
		{
			free(p1);
			p1 = NULL;
		}
		else if (p1 && p2)
		{
			tmp = p1;
			tmp->next = NULL;
			free(tmp);
			p1 = p2;
			p2 = p2->next;
		}
	}
}

char *getname(unsigned long addr)
{
	struct hostent *h;

	h = gethostbyaddr((char *)&addr, 4, AF_INET);
	return h->h_name;
}

void dump_status()
{
	stream_t *s1;
	client_t *c1;

	printf("\n The Status of the streams & clients :=>\n");


	for (s1=stream_list; s1!=NULL; s1=s1->next)
	{
		if (s1->movie_id!=-1)
		{
			printf("Stream ID: %x. Movie ID: %d (%s). Clients:  ", 
						 s1->stream_id, s1->movie_id, 
						 s1->movie_id==-1?"Disabled":movielist[s1->movie_id]);
			for (c1 = s1->clients; c1!=NULL; c1=c1->next)
				printf("%s ", getname(c1->hostid));
			printf("\n");
		}
	}
}


void request_movie(int movie_id, unsigned long hostid)
{
	stream_t *s1, *s2, i;

	if ((s1 = find_client(hostid))!=NULL)
	{
/* 		err_msg("Found client in stream %d", s1->stream_id); */
		remove_client_from_stream(s1, hostid);
/* 		err_msg("Client removed"); */
		if (s1->clients==NULL)
		{
/* 	    err_msg("Disabling stream %d ", s1->stream_id); */
	    disable_stream(s1->stream_id);
			/*remove_stream(s1->stream_id);*/
			dump_status();
		}
	}

	for (s1=stream_list; s1!=NULL; s1=s1->next)
		if (s1->movie_id == movie_id)
	    break;

	if (s1==NULL) /* Normal stream */
	{
		s1 = create_stream(movie_id, 0);
/* 		err_msg("Stream %d created", s1->stream_id); */
	}
	else   /* Accelerated stream */
	{
		movie_id += num_movies/2; /* Accelerated movies are kept at the end of the list */
/* 		err_msg("Accelerated movie id %d\n", movie_id); */
		s1 = create_stream(movie_id, 0);
/* 		err_msg("Accelerated stream %d created", s1->stream_id);  */
	}
	
	add_client_to_stream(s1, hostid);
 	printf("Added client %d to stream %x", hostid, s1->stream_id); 
	issue_request(movielist[movie_id], s1->stream_id, htons(VIDEO_PORTNUM));

}

void setup_socket()
{
	struct sockaddr_in addr;
	int addrlen;

	if ((sockfd = socket(AF_INET, SOCK_DGRAM, 0)) <0)
		err_sys("Unable to create a socket");

	bzero((char *) &addr, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = htonl(INADDR_ANY);
	addr.sin_port = htons(DSAP_PORTNUM);
	addrlen = sizeof(addr);
    
	if (bind(sockfd, (struct sockaddr *) &addr, addrlen)<0)
		err_sys("Unable to bind");

}


void merge_streams()
{
	int *pid_val;
	client_t *c;
	stream_t *sa, *sn;
	unsigned long normal_group, accel_group;
		
	if((shm_for_pid = shmget(SHMKEY0, (NSERVERS+1)*sizeof(int), PERMS)) < 0)
	{
		err_sys("Could not get shared memory");
	}

	if((pid_val = (int *) shmat(shm_for_pid, (char *) 0, 0)) == (int *)-1)
	{
		err_sys("Could not attach shared memory");
	}
	
	sem_wait(sem_for_pid);     /* Lock shared memory for write */

	normal_group = *(pid_val+1);
/* 	err_msg("Normal Group %d", normal_group); */
	accel_group = *(pid_val+2);
/* 	err_msg("Accelerated Group %d", accel_group); */

	*pid_val = 1;
	bzero(pid_val+1, NSERVERS);
	
	sem_signal(sem_for_pid);   /* Unlock shared memory */

	if(shmdt((void *) pid_val) < 0)
	{
		err_sys("Could not detach shared memory");
	}

	sa = find_stream(accel_group);
	sn = find_stream(normal_group);
	
	for (c=sa->clients; c!=NULL; c = sa->clients)
	{
		send_to_client(c->hostid, CC_SWITCH_CHANNEL, normal_group);
		add_client_to_stream(sn, c->hostid);
		remove_client_from_stream(sa, c->hostid); 
	}

/* 	err_msg("Before remove stream"); */

	disable_stream(accel_group);
	remove_stream(accel_group);
	dump_status();

/* 	err_msg("After remove stream"); */
}

void handle_pause_request(stream_t *st, unsigned long client_id)
{
	client_t *c = st->clients;
	stream_t *new_st = NULL;
	
	if (c->next != NULL) /* there are more than one client with that stream */
	{
		remove_client_from_stream(st, client_id);
		new_st = create_stream(st->movie_id + (num_movies/2), st->frame_num);
		add_client_to_stream(new_st, client_id);
		dump_status();
/* 		err_msg("Sending message to client"); */
		send_to_client(client_id, CC_SWITCH_CHANNEL, new_st->stream_id);

		/* Now enter into the shared memory */
		enter_pause(st->stream_id, new_st->stream_id);
	}
	else /* there is just one client with that stream */
	{
		mark_entry(st->stream_id, S_ACTIVE, S_PAUSED);
/* 		err_msg("Sending message to client"); */
		send_to_client(client_id, CC_SWITCH_CHANNEL, st->stream_id);
	}
}


void handle_quit_request(stream_t *st, unsigned long client_id)
{
	client_t *c = st->clients;
	
	if (c->next != NULL) /* there are more than one client with that stream */
	{
		remove_client_from_stream(st, client_id);
		dump_status();
	}
	else /* there is just one client with that stream */
	{
		remove_client_from_stream(st, client_id);
		disable_stream(st->stream_id);
		/*		remove_stream(st->stream_id);*/
		dump_status();
	}
}



void accept_requests()
{
	int n, cli_len;
	unsigned long i, j;
	unsigned char data[8];
	struct sockaddr_in cli;
	stream_t *st;
	
	cli_len = sizeof(cli);

	for (; ;) 
	{
		bzero(&cli, cli_len);
/* 		err_msg("Client before --> %d", cli.sin_addr.s_addr); */
		n = recvfrom(sockfd, data, 8, 0, (struct sockaddr *) &cli, &cli_len);		
/* 		err_msg("Client after --> %d, Data %d", cli.sin_addr.s_addr, data[0]); */

		if (n < 0)
		{
	    err_msg("VCR: recvfrom error: Error number %d", errno);
			if (errno == EINTR)
				continue;
			else err_sys("Fatal error");
		}
		
		switch(data[0])
		{
		case CC_START_MOVIE:  /* from client */
	    n = BYTES2SHORT(data+1);
	    i = cli.sin_addr.s_addr;
	    request_movie(n, i);
	    break;
			
		case CC_PAUSE:   /* from client */
			i = cli.sin_addr.s_addr;
			st = find_client(i);

			if (st != NULL) 
				handle_pause_request(st, i);

			break;

		case CC_UNPAUSE:   /* from client */
			i = cli.sin_addr.s_addr;
			st = find_client(i);
			i = st->stream_id;

			mark_entry(i, S_PAUSED, S_ACTIVE);			
	    break;

		case CC_MERGED: /* Got from the server process */
			merge_streams();
			break;
			
		case CC_QUIT:   /* from client */
			i = cli.sin_addr.s_addr;
			st = find_client(i);

			if (st != NULL)
				handle_quit_request(st, i);			

			break;			

		case CC_SERVER_SIGNAL:   /* from the server process when the clip ends */
			i = BYTES2LONG(data+1);
			remove_all_clients_from_stream(i);
			remove_stream(i);
			break;
		}

		dump_status();

	}
			
}


void main(int argc, char *argv[]) 
{
	int                     c, debug = 0;
	char                    *pname = argv[0];
	
	if((fperr=fopen(ERRORLOG, "a"))==NULL) {
		perror("Unable to open Error Log File");
	}

	while((c = getopt(argc, argv, "d")) != EOF) {
		switch(c) {
		case 'd':                                 /* debug */
	    debug = 1;
	    break;
		case '?':
	    err_quit("%s: usage: %s [ -d ]", pname, pname);
		}
	}

	convey_pid(); /* Convey own pid to the server process */
	
	stream_list = NULL;
	start_daemons(debug);
	err_msg("DSAP: Daemon started\n"); 

	read_movies();

	setup_socket();
/* 	err_msg("DSAP: Socket set up\n"); */

	accept_requests();

	fclose(fperr);
}


