#include "netdefs.h"
#include "path.h"
#include "vcripc.h"
#include "vcrctrl.h"
#include "server.h"
#include "cchannel.h"
#include "dsap.h"
#include "issue.h"

static short stat_id = 1;

void issue_request(char * name, unsigned long multicast_grp,unsigned short portnum)
{
	LOGINTABLE *loginrec, *firstrec;
	int semaphid, shmid, i;
     
	logmessage("ISSUEREQ", "INF", "issuing a request");

	if((shmid = shmget(SHMKEY, NSERVERS*sizeof(LOGINTABLE), 
										 PERMS)) < 0)
	{
		logmessage("ISSUEREQ", "ERR", "can't get shared memory");
		exit(1);
	}     

	if((semaphid = sem_open(SEMKEY)) < 0)
	{
		logmessage("ISSUEREQ", "ERR", "can't open semaphore");
		exit(1);
	}    

	if((firstrec = (LOGINTABLE *)shmat(shmid, (char *) 0, 0)) 
		 == (LOGINTABLE *)-1)
	{
		logmessage("ISSUEREQ", "ERR", "Can't attach shared memory");
		exit(1);
	}
	/* Create entry in table */
	sem_wait(semaphid);     /* Lock shared memory for write */

	/*
	for (i=0, loginrec = firstrec; i<NSERVERS; i++, loginrec++)
		printf ("\n Rec here %d %d %d %d\n", loginrec->status, loginrec->ID, 
						loginrec->multiaddr, loginrec->currframe);
						*/
	
	if((loginrec = find_empty(firstrec)) == NULL)
	{
		logmessage("ISSUEREQ", "ERR", "No free entries");
		exit(1);
	}
	logmessage("ISSUEREQ", "INF", "Setting as active");

	loginrec->status = S_ACTIVE | S_NEWREQUEST;
	strcpy(loginrec->filename,name);
	loginrec->multiaddr = multicast_grp;
	loginrec->portnum = portnum;
	loginrec->ID = stat_id++;
	loginrec->currframe = 0;
	loginrec->filepos = 0;
	
	logmessage("ISSUEREQ", "INF", "Got user request");

	/*
	for (i=0, loginrec = firstrec; i<NSERVERS; i++, loginrec++)
		printf ("\n Rec here %d %d %d %d\n", loginrec->status, loginrec->ID, 
						loginrec->multiaddr, loginrec->currframe);
						*/

	sem_signal(semaphid);   /* Unlock shared memory */

	if(shmdt((void *)firstrec) < 0)
	{
		logmessage("ISSUEREQ", "ERR", "Can't detach shared mem");
		exit(1);
	}
}

int remove_entry(int mcast_group)
{
	LOGINTABLE *loginrec, *firstrec;
	int semaphid, shmid;
	int retval;
	int i;

	logmessage("ISSUEREQ", "INF", "issuing a request");

	if((shmid = shmget(SHMKEY, NSERVERS*sizeof(LOGINTABLE), 
										 PERMS)) < 0)
	{
		logmessage("ISSUEREQ", "ERR", "can't get shared memory");
		exit(1);
	}     

	if((semaphid = sem_open(SEMKEY)) < 0)
	{
		logmessage("ISSUEREQ", "ERR", "can't open semaphore");
		exit(1);
	}    

	if((firstrec = (LOGINTABLE *)shmat(shmid, (char *) 0, 0)) 
		 == (LOGINTABLE *)-1)
	{
		logmessage("ISSUEREQ", "ERR", "Can't attach shared memory");
		exit(1);
	}
	/* Create entry in table */
	sem_wait(semaphid);     /* Lock shared memory for write */

	retval = 0;
	for (i=0, loginrec = firstrec; i<NSERVERS; i++, loginrec++)
	{

		printf ("Loginrec %d %d %d Mcast_group %d\n", loginrec->status, loginrec->ID, 
						loginrec->multiaddr, mcast_group);
		if (loginrec->multiaddr==mcast_group)
		{
			/* delete this entry */
			logmessage("REMOVEENTRY", "INF", "Removing request ");
			bzero(loginrec, sizeof(LOGINTABLE));
			retval = 1;
			break;
		}
	}
	sem_signal(semaphid);   /* Unlock shared memory */
	if(shmdt((void *)firstrec) < 0)
	{
		logmessage("ISSUEREQ", "ERR", "Can't detach shared mem");
		exit(1);
	}

	return retval;
}

void mark_entry(unsigned long mcast_group, unsigned short old_status, unsigned short new_status)
{
	LOGINTABLE *loginrec, *firstrec;
	int semaphid, shmid;
	int i;

	logmessage("ISSUEREQ", "INF", "issuing a request");

	if((shmid = shmget(SHMKEY, NSERVERS*sizeof(LOGINTABLE), 
										 PERMS)) < 0)
	{
		logmessage("ISSUEREQ", "ERR", "can't get shared memory");
		exit(1);
	}     

	if((semaphid = sem_open(SEMKEY)) < 0)
	{
		logmessage("ISSUEREQ", "ERR", "can't open semaphore");
		exit(1);
	}    

	if((firstrec = (LOGINTABLE *)shmat(shmid, (char *) 0, 0)) 
		 == (LOGINTABLE *)-1)
	{
		logmessage("ISSUEREQ", "ERR", "Can't attach shared memory");
		exit(1);
	}
	/* Create entry in table */
	sem_wait(semaphid);     /* Lock shared memory for write */

	for (i=0, loginrec = firstrec; i<NSERVERS; i++, loginrec++)
	{
/* 		err_msg("Loginrec: status:%d ID:%d port:%d Mcast_group:%d", loginrec->status,  */
/* 						loginrec->ID, loginrec->portnum, loginrec->multiaddr); */
		if ((loginrec->status & old_status) && (loginrec->multiaddr==mcast_group))
		{
/* 			err_msg("Marking entry as %d", new_status); */
			loginrec->status |= new_status;
			loginrec->status &= (0xff - old_status);
			break;
		}
	}
	sem_signal(semaphid);   /* Unlock shared memory */

	if(shmdt((void *)firstrec) < 0)
	{
		logmessage("ISSUEREQ", "ERR", "Can't detach shared mem");
		exit(1);
	}
}



void enter_pause(unsigned long old_mcast_group, unsigned long new_mcast_group)
{
	LOGINTABLE *loginrec, *firstrec, *newrec;
	int semaphid, shmid;
	int i;

	if((shmid = shmget(SHMKEY, NSERVERS*sizeof(LOGINTABLE), PERMS)) < 0)
	{
		err_sys("Can't get shared memory");
	}     

	if((semaphid = sem_open(SEMKEY)) < 0)
	{
		err_sys("Can't open semaphore");
	}    

	if((firstrec = (LOGINTABLE *)shmat(shmid, (char *) 0, 0)) 
		 == (LOGINTABLE *)-1)
	{
		err_sys("Can't attach shared memory");
	}

	/* Create a new entry in table */
	sem_wait(semaphid);     /* Lock shared memory for write */

	loginrec = find_entry(firstrec, old_mcast_group);
	newrec = find_empty(firstrec);

	newrec->ID = stat_id++;
	newrec->status = loginrec->status;
	newrec->status |= S_PAUSED;
	newrec->status &= (0xff - S_ACTIVE);
	newrec->multiaddr = new_mcast_group;
	newrec->portnum = loginrec->portnum;
	newrec->currframe = (14*loginrec->currframe)/15;
	/*newrec->filepos = loginrec->filepos;*/
	strcpy(newrec->filename, loginrec->filename);
	strcat(newrec->filename, "a");
	
/* 	err_msg("Loginrec: status:%d ID:%d filename:%s Mcast_group:%d Filepos:%d", loginrec->status,  */
/* 						loginrec->ID, loginrec->filename, loginrec->multiaddr, loginrec->filepos); */

/* 	err_msg("Newrec: status:%d ID:%d filename:%s Mcast_group:%d Filepos:%d", newrec->status,  */
/* 					newrec->ID, newrec->filename, newrec->multiaddr, newrec->filepos); */

	sem_signal(semaphid);   /* Unlock shared memory */

	if(shmdt((void *)firstrec) < 0)
	{
		err_sys("Can't detach shared memory");
	}

}



void convey_pid()
{
	int *pid_val;
	
	if((shm_for_pid = shmget(SHMKEY0, (NSERVERS+1)*sizeof(int), PERMS|IPC_CREAT)) < 0)
	{
		err_sys("Could not get shared memory");
	}

	if((sem_for_pid = sem_create(SEMKEY0, 1)) < 0)
	{
		err_sys("Could not create semaphore for pid");
	}


	if((pid_val = (int *) shmat(shm_for_pid, (char *) 0, 0)) == (int *)-1)
	{
		err_sys("Could not attach shared memory");
	}
	
	if((sem_for_pid = sem_open(SEMKEY0)) < 0)
	{
		err_sys("Could not open semaphore");
	}

	sem_wait(sem_for_pid);     /* Lock shared memory for write */
	
	bzero(pid_val, NSERVERS);

	*pid_val = getpid();
/* 	err_msg("\t\t\t PID val %d\n", *pid_val); */

	sem_signal(sem_for_pid);   /* Unlock shared memory */

	if(shmdt((void *) pid_val) < 0)
	{
		err_sys("Could not detach shared memory");
	}


}

