#include "netdefs.h"
#include "vcripc.h"

/*******************************************************************/
/*                                                                 */
/*    Description      : This function returns the pointer to the  */ 
/*                       first free record ( i.e. where b_active   */
/*                       is zero) in the login table.              */
/*                                                                 */
/*******************************************************************/
LOGINTABLE *find_empty(LOGINTABLE *firstrec)
{
     LOGINTABLE *thisrec;
		 int i;

     /* Seek beyond last record */
		 thisrec = firstrec;

     for (i=0; i<NSERVERS; i++)
     {
          /* If not marked active, return a pointer */
          if(!(thisrec->status & (S_ACTIVE | S_PAUSED | S_STOPPING | S_STOPREQUEST))) return(thisrec);
					thisrec ++;
     } 

     return((LOGINTABLE *)NULL); 
}/* End of find_empty() */


/*******************************************************************/
/*                                                                 */
/*    Description      : This function returns a pointer to the    */ 
/*                       the active record in the logins table     */
/*                       where pid equals the specified pid.       */
/*                                                                 */
/*******************************************************************/
LOGINTABLE *find_pid(LOGINTABLE *firstrec, unsigned long ID)
{
     LOGINTABLE *thisrec;
		 int i;
		 
     thisrec = firstrec;

     for (i=0; i<NSERVERS; i++)
     {
			 /* If marked active, Check for a match */
			 if((thisrec->status & (S_ACTIVE | S_PAUSED | S_STOPPING | S_STOPREQUEST)) && (thisrec->ID == ID)) 
				 return(thisrec);
			 thisrec++;
     } 

     return((LOGINTABLE *)NULL); 
}/* End of find_pid() */



LOGINTABLE *find_entry(LOGINTABLE *firstrec, unsigned long mcast_group)
{
     LOGINTABLE *thisrec;
		 int i;
		 
     thisrec = firstrec;

     for (i=0; i<NSERVERS; i++)
     {
			 /* If marked active, Check for a match */
			 if((thisrec->status & (S_ACTIVE | S_PAUSED | S_STOPPING | S_STOPREQUEST)) && (thisrec->multiaddr == mcast_group)) 
				 return(thisrec);
			 thisrec++;
     } 

     return((LOGINTABLE *)NULL); 
}/* End of find_entry() */
