#include "netdefs.h"
#include "vcrctrl.h"
#include "vcripc.h"
#include "server.h"
#include "path.h"
#include "error.h"
#include "issue.h"

extern int errno;
FILE *fperr;                 /* Global error log file */
int semaphid;
int shmid;

void shut_down();

void logmessage(char *c_facility, char *c_type, char *c_msg)
{
	time_t t;

	t = time(&t);
	if(! LOGLEVEL) return;
	fprintf(fperr,"%s%%%s::%s @ %s", c_facility, c_type, c_msg, ctime(&t)); 
	fflush(fperr);
}

void create_ipcs(void)
{
	LOGINTABLE *loginrec;

	/* Create shared memory segment */
	if((shmid = shmget(SHMKEY, NSERVERS*sizeof(LOGINTABLE), 
										 PERMS|IPC_CREAT)) < 0)
	{
		logmessage("VCRSERVD", "ERR", "can't get shared memory");
		exit(1);
	}

	/* Create the semaphore. */
	if((semaphid = sem_create(SEMKEY, 1)) < 0)
	{
		logmessage("VCRSERVD", "ERR", "can't create semaphore");
		exit(1);
	}    

	/* Gotta clear the table here */
	if((loginrec = (LOGINTABLE *)shmat(shmid, (char *) 0, 0))
		 == (LOGINTABLE *)-1)
	{
		logmessage("VCRSERVD", "ERR", "can't attach shared memory");
		exit(1);
	}

	bzero(loginrec, NSERVERS*sizeof(LOGINTABLE));

	if(shmdt((void *)loginrec) < 0)
	{
		logmessage("VCRSERVD", "ERR", "can't detach shared memory");
		exit(1);
	}

}

void remove_ipcs(void)
{
	if(shmctl(shmid, IPC_RMID, (struct shmid_ds *) 0) < 0)
		logmessage("VCRSERVD", "ERR", "can't remove shared memory");

	if(shmctl(shm_for_pid, IPC_RMID, (struct shmid_ds *) 0) < 0)
		logmessage("VCRSERVD", "ERR", "can't remove shared memory");

	logmessage("VCRSERVD", "INF", "removed shared memory");

	sem_rm(semaphid);
	sem_rm(sem_for_pid);
}

void shut_down()
{
	logmessage("VCRSERV", "INF", "trying to shutdown");
	remove_ipcs();
	logmessage("VCRSERVD", "INF", "shutting down");
#ifdef LINUX
	signal(SIGABRT, SIG_DFL);
#endif
	abort();
}

void daemon_start(void)
{
	register int i;
	register int childpid;

	if((fperr=fopen(ERRORLOG, "a"))==NULL)
	{
		fprintf(stderr,"can't open error log file \n");
		fflush(stderr);
		exit(1);
	}

	logmessage("VCRSERVD", "INF", "Attempting to start daemon");

	/* Ignore all signals */
	for(i=0; i<33; i++) {
		if( (i != SIGTERM || i!=SIGINT))
			signal(i, SIG_IGN);
	}
	errno=0;

	signal(SIGTERM, SIG_DFL);
	signal(SIGINT, SIG_DFL);
	signal(SIGTERM, shut_down);
	signal(SIGINT, shut_down);

	if((childpid=fork()) < 0)
	{
		logmessage("VCRSERVD", "ERR", "Can't fork first child for daemon");
		exit(1);                          /* if not started in background */
		/* fork and let parent exit    */
	}
	else if(childpid>0) 
		exit(0);           /* parent */

	/* First child process */
	if(setsid() == -1) 
	{
		logmessage("VCRSERVD", "ERR", "Can't change process group");
		exit(1);
	}

	logmessage("VCRSERVD", "INF", "Before second fork");
	if((childpid=fork())<0)
	{
		logmessage("VCRSERVD", "ERR", "Can't fork second child for daemon");
		exit(1);
	}
	else if(childpid > 0) 
		exit(0);            /* first child */

	/* Second child process */
	logmessage("VCRSERVD", "INF", "Daemon start successful");

	/* Close all files */
	for(i=0;i<NOFILE;i++) close(i);
	errno=0;

	chdir("/");
	umask(0);
}


int start_daemons(int debug)
{

  if(debug == 0)  
		daemon_start();  

	create_ipcs();  

	atexit(shut_down);     
	signal(SIGTERM,shut_down);
  
	logmessage("VCRSERVD", "INF", "Starting VCR Server");
}






