/************************************************************

	File: cchannel.h

	Author: Ashok

	Purpose: This file defines the control channel interface. Remember
	         that the other end of the control channel is written in Java,
		 so this file is only used by the DSAP agent. Also, we must adhere
		 strictly to a byte-level protocol only (no ints)

*************************************************************************/

#if !defined(_CCHANNEL)
#define _CCHANNEL

/* The first byte sent across the channel represents a message type code.
	 These are the possible codes that can be sent.

	 CC_START_MOVIE <short movie_id> : client-to-agent
	 CC_SWITCH_CHANNEL <unsigned long mcast_addr> : agent-to-client
	 CC_PAUSE: client-to-agent
	 CC_UNPAUSE: client-to-agent
	 CC_QUIT: client-to-agent

	 The next few are server-agent communication
	 
	 CC_SERVER_SIGNAL : signal from server to agent that a cycle is complete

	 */

#define LONG2BYTES(iv, cp)  (cp)[0] = (iv) >> 24;\
                            (cp)[1] = ((iv) >> 16) & 0xFF;\
                            (cp)[2] = ((iv) >> 8) & 0xFF;\
                            (cp)[3] = (iv) & 0xFF;

#define SHORT2BYTES(iv, cp) (cp)[0] = (iv) >> 8;\
                            (cp)[1] = (iv) & 0xFF;
 
#define BYTES2SHORT(cp) ((short)((((int)((cp)[0]))<<8) + ((int)((cp)[1]))))

#define BYTES2LONG(cp) ((long)((((int)((cp)[0]))<<24) + (((int)((cp)[1]))<<16) + (((int)((cp)[2]))<<8) + ((int)((cp)[3]))))

#define CC_START_MOVIE         1   
#define CC_SWITCH_CHANNEL      2
#define CC_PAUSE               3
#define CC_UNPAUSE             4
#define CC_QUIT                5
#define CC_SERVER_SIGNAL       6
#define CC_MERGED              7

#define PACKET_SIZE  536

#endif
