#include "netdefs.h"
#include "error.h"
#include "vcripc.h"
#include "newsess.h"
#include "vcrctrl.h"
#include "util.h"
#include "path.h"
#include <errno.h>

#include "cchannel.h"

#define FRAMES_PER_SEC  30


extern int fildes, errno;
extern char *pname;
extern int dsap_pid;
const int addrsize = sizeof(struct sockaddr_in); 

unsigned char *largebuffer, *pkt_ptr;
extern int sem_for_pid, shm_for_pid;

static struct sess_info *first = NULL;  /* pointer to the first session */
static int sess_no = 0;                 /* current number of sessions */
static int sockfd,                      /* UDP socket - DATA socket */
	   semaphid,                    /* semaphore id */
	   shmid;                       /* shared memory segment id */

int blocknum;

unsigned long frameOff[MAX_NUM_OF_FRAMES];   /* the frame offsets in the conditioned file */

/*
 * Initializes server's parameters.
 */
void init_serv(void) {
	int i;


	/* Open shared memory */
	if((shmid = shmget(SHMKEY, NSERVERS*sizeof(LOGINTABLE), 0)) < 0) {
		printf("%d\n", errno);
		err_sys("ERROR: cannot get shared memory");
	}
     
	/* Open the semaphore */
	if((semaphid = sem_open(SEMKEY)) < 0)
		err_sys("ERROR: cannot open the semaphore");
}

/*
 * Checks shared memory table. Table record (LOGINTABLE) defined
 * in vcripc.h
 */
void check_shmem() {
	LOGINTABLE *firstrec, *thisrec;
	struct sess_info *tmp;
	int i;

	/* Attach shared memory segment */
	if((firstrec = (LOGINTABLE *) shmat(shmid, (char *) 0, 0)) 
		 == (LOGINTABLE *) -1)
		err_sys("ERROR: cannot attach shared memory");

	sem_wait(semaphid);                     /* lock shared memory */
	for(i=0; i<NSERVERS; i++) {
		thisrec = firstrec + i;
		if(thisrec->status & S_ACTIVE) {                 /* check if session is active */

	    if(thisrec->status & S_NEWREQUEST) {     /* check if session is new */
				thisrec->status &= (0xff - S_NEWREQUEST);
				thisrec -> status |= S_STARTING;
				init_sess(thisrec); /* create new session */
				continue;
	    }

	    if (thisrec ->status & S_STARTING)
	    {
				thisrec -> status &= (0xff - S_STARTING);
				init_sess(thisrec);
				continue;
	    }

	    tmp = first;
	    while (tmp->id != thisrec->ID)
	    {				
				tmp = tmp->next;
				if (tmp==NULL)
					err_sys("Blown a fuse here!");
	    }
	    tmp->status = thisrec -> status;
			err_msg("%d: ACTIVE=> id:%d currframe:%d filepos:%d status:%d filename:%s", getpid(),
							tmp->id, thisrec->currframe, thisrec->filepos, tmp->status, thisrec->filename);

		}	


		if (thisrec->status & S_PAUSED) { /* Just copy the status info to the session ptrs */	    
	    tmp = first;
	    while (tmp->id != thisrec->ID)
	    {
				tmp = tmp->next;
				if (tmp==NULL)
				{
					create_sess(thisrec);
					err_msg("%d: PAUSE=> id:%d currframe:%d filepos:%d status:%d filename:%s", getpid(),
									first->id, thisrec->currframe, thisrec->filepos, first->status, thisrec->filename);
					break;
				}	
	    }
			if (tmp)
				tmp->status = thisrec -> status;
		}


		if(thisrec->status & S_STOPPING) {      
	    remove_sess(thisrec->ID);       /* request is stop */
	    err_msg("STOPPING: Remove session called");
	    thisrec->status &= (0xff-S_STOPPING);
	    bzero(thisrec, sizeof(LOGINTABLE));
	    thisrec->status |= S_STOPPED;
			continue;
		}

		if(thisrec->status & S_STOPREQUEST) {
	    remove_sess(thisrec->ID);       /* request is stop */
	    err_msg("%d: STOPREQUEST: Remove session called", getpid()); 
	    thisrec ->status &= (0xff - S_STOPREQUEST);
	    thisrec -> status |= S_STOPPING;
			err_msg("%d: STOPREQUEST: Made it Stopping: Status %d", getpid(), thisrec->status);
	    continue;
		}
	}

	/* Rate adaptation Code */
	rate_adapt(firstrec);
	
	sem_signal(semaphid);                   /* unlock shared memory */
     
	/* Detach shared memory segment */
	if(shmdt((void *)firstrec) < 0)
		err_sys("ERROR: cannot detach shared memory");
}
 
void rate_adapt(LOGINTABLE *firstrec)
{
	int i, j, *pid_val;
	LOGINTABLE *rec1, *rec2;

	err_msg("%d: In rate_adapt()", getpid());

	if((shm_for_pid = shmget(SHMKEY0, (NSERVERS+1)*sizeof(int), PERMS)) < 0)
	{
		err_sys("Could not get shared memory");
	}

	if((pid_val = (int *) shmat(shm_for_pid, (char *) 0, 0)) == (int *)-1)
	{
		err_sys("Could not attach shared memory");
	}
	
		
	sem_wait(sem_for_pid);

	for (i = 0; i < NSERVERS; i++)
	{
		rec1 = firstrec + i;
		for (j = i+1; j < NSERVERS; j++)
		{
			rec2 = firstrec + j;

			if ((rec2->status & S_ACTIVE) && (rec1->status & S_ACTIVE) && *pid_val)
			{
				if 	((strstr(rec2->filename, rec1->filename)) && ((rec1->currframe/16) == ((rec2->currframe/15))))
				{
					enter_stream_ids(pid_val+1, rec1->multiaddr, rec2->multiaddr);
					*pid_val = 0;
				}
				else if ((strstr(rec1->filename, rec2->filename)) && ((rec2->currframe/16) == ((rec1->currframe/15))))
				{
					enter_stream_ids(pid_val+1, rec2->multiaddr, rec1->multiaddr);
					*pid_val = 0;
				}
			}
		}
	}

	if (*pid_val == 0)
	{
		struct sockaddr_in addr;
		unsigned char c = CC_MERGED;
		int sock;

		sock = socket(AF_INET, SOCK_DGRAM, 0);
		if (sock<0)
			err_sys("Can't create socket");

		bzero(&addr, sizeof(addr));
		addr.sin_addr.s_addr = htonl(INADDR_ANY);
		addr.sin_port = htons(DSAP_PORTNUM);
		addr.sin_family = AF_INET;

		if (sendto(sock, &c, 1, 0, (struct sockaddr *)&addr,sizeof(addr))<0)
			err_sys("Can't send UDP packet");
	}
		
		
	
	sem_signal(sem_for_pid); 
	

	if(shmdt((void *) pid_val) < 0)
	{
		err_sys("Could not detach shared memory");
	}
}

void enter_stream_ids(int *pid_val, unsigned long stream_id1,  unsigned long stream_id2)
{
	int i, *ptr;

	ptr = pid_val;

	if (*ptr == 0)                    /* first entry Zero  */
	{
		*ptr++ = stream_id1;
		*ptr = stream_id2;
	}
}


/* For paused sessions */
void create_sess(LOGINTABLE *rec) {
	LOGINTABLE *loginrec;
	char fname[MAXLINE], fname2[MAXLINE];
	struct sess_info *ptr;
	int j, i;
 
	/* Allocate memory for session info, buffer and header */
	if((ptr = malloc(sizeof(struct sess_info))) == NULL)
		err_sys("ERROR: cannot allocate memory");

	for (j=0; j<15; j++) {
		if((ptr->frbuf[j] = (BYTE *) malloc(sizeof(BYTE) * BUFSIZE)) == NULL)
	    err_sys("ERROR: cannot allocate memory");
	}
	/* Copy session info from the shared memory table */
     
	/* Open file to read from */
	sprintf(fname, "%s%s", VPATH, rec->filename);     
	if((ptr->fd = open(fname, O_RDONLY)) < 0)
		err_sys("ERROR: cannot open %s",fname);  


	/* Open the frame offsets file */
	sprintf(fname2, "%s%s_frame", VPATH, rec->filename);     
	if((ptr->fdframe = open(fname2, O_RDONLY)) < 0)
		err_sys("ERROR: cannot open %s",fname2);  

	ptr->num_of_frames = READ_INT(ptr->fdframe);
	ptr->currframe = rec->currframe;
	ptr->on = 1;

	ptr->status = rec->status;
	
	if((ptr->file_len = filelen(ptr->fd)) == -1)
		err_sys("ERROR: cannot compute length of %s",fname);

	/* Create what looks like an ordinary UDP socket */
	if ((sockfd = socket(AF_INET,SOCK_DGRAM,0)) < 0) 
		err_sys("ERROR: cannot create socket");
        
	/* Set up destination address */    
	memset(&(ptr->to), 0, addrsize); 
	(ptr->to).sin_family        = AF_INET; 
	(ptr->to).sin_addr.s_addr   = rec->multiaddr; 
	(ptr->to).sin_port          = rec->portnum; 
    
	/* Give session a unique id */
	ptr->id = rec->ID;
	
	/* Update linked list of sessions */
	ptr->next = first;
	first = ptr;
	sess_no++;        
       
	err_msg("Number Of sessions: %d", sess_no);
}



/*
 * Initializes new session.
 * Takes pointer to the table record.
 */
void init_sess(LOGINTABLE *rec) {
	LOGINTABLE *loginrec;
	char fname[MAXLINE], fname2[MAXLINE];
	struct sess_info *ptr;
	int j, i;
 
	/* Allocate memory for session info, buffer and header */
	if((ptr = malloc(sizeof(struct sess_info))) == NULL)
		err_sys("ERROR: cannot allocate memory");

	for (j=0; j<15; j++) {
		if((ptr->frbuf[j] = (BYTE *) malloc(sizeof(BYTE) * BUFSIZE)) == NULL)
	    err_sys("ERROR: cannot allocate memory");
	}
	/* Copy session info from the shared memory table */
     
	/* Open file to read from */
	sprintf(fname, "%s%s", VPATH, rec->filename);     
	if((ptr->fd = open(fname, O_RDONLY)) < 0)
		err_sys("ERROR: cannot open %s",fname);  


	/* Open the frame offsets file */
	sprintf(fname2, "%s%s_frame", VPATH, rec->filename);     
	if((ptr->fdframe = open(fname2, O_RDONLY)) < 0)
		err_sys("ERROR: cannot open %s",fname2);  

	ptr->num_of_frames = READ_INT(ptr->fdframe);

	ptr->currframe = 0;
	
	ptr->on = 1;

	/* Compute the length of the file (in bytes) */
	if((ptr->file_len = filelen(ptr->fd)) == -1)
		err_sys("ERROR: cannot compute length of %s",fname);

	/* Create what looks like an ordinary UDP socket */
	if ((sockfd = socket(AF_INET,SOCK_DGRAM,0)) < 0) 
		err_sys("ERROR: cannot create socket");
        
	/* Set up destination address */    
	memset(&(ptr->to), 0, addrsize); 
	(ptr->to).sin_family        = AF_INET; 
	(ptr->to).sin_addr.s_addr   = rec->multiaddr; 
	(ptr->to).sin_port          = rec->portnum; 
    
	/* Give session a unique id */
	ptr->id = rec->ID;
	
	err_msg("pid %d: multicast session started @ %sReading from file %s",  
					getpid(), tstamp(), fname);  

	/* Update linked list of sessions */
	ptr->next = first;
	first = ptr;
	sess_no++;        
       
	err_msg("Number Of sessions: %d", sess_no);
}

/*
 * Finds and removes session with the given id from the linked list.
 * Takes session id.
 */
void remove_sess(unsigned long sess_id) {
  int j;
  struct sess_info *ptr, *prev;

  ptr = prev = first;

	if(ptr == NULL) {
    err_msg("%s: ERROR: session %d not found", pname, sess_id);
    return;
  }
	
  if(ptr->id == sess_id) {
    first = ptr->next;

    for (j=0; j<15; j++) 
      free(ptr->frbuf[j]);
		
		free(ptr);
		
    sess_no--;
    err_msg("%s: pid %d: session %d terminated @ %sNumber of sessions: %d",
						pname, getpid(), sess_id, tstamp(), sess_no);
    return;
  }
  ptr = ptr->next;
  while(ptr != NULL) {
    if(ptr->id == sess_id) {
      prev->next = ptr->next;

      for (j=0; j<15; j++)
				free(ptr->frbuf[j]);
			
      free(ptr);
		
      sess_no--;
      err_msg("%s: pid %d: session %d terminated @ %sNumber of sessions: %d",
							pname, getpid(), sess_id, tstamp(), sess_no);
      return;
    }
    prev = ptr;
    ptr = ptr->next;
  }
  err_msg("%s: ERROR: session %d not found", pname, sess_id);
}




/* At EOF - issue a request for cleanup to shared memory */
void server_cleanup(unsigned long id) {
	LOGINTABLE *firstrec, *thisrec;

	/* Attach shared memory segment */
	if((firstrec = (LOGINTABLE *) shmat(shmid, (char *) 0, 0)) 
		 == (LOGINTABLE *) -1)
		err_sys("ERROR: cannot attach shared memory");

	sem_wait(semaphid);                     /* lock shared memory */

	/* Find correct login record */
	thisrec = find_pid(firstrec,id);

	if (thisrec) {
		if  (thisrec->status & S_ACTIVE) 
		{
			struct sockaddr_in addr;
			unsigned char c[5];
			int sock;

			thisrec->status |= S_STOPREQUEST;
			thisrec->status &= (0xff - S_ACTIVE);

			/* Send message to the DSAP agent */
			sock = socket(AF_INET, SOCK_DGRAM, 0);
			if (sock<0)
				err_sys("Can't create socket");

			bzero(&addr, sizeof(addr));
			addr.sin_addr.s_addr = htonl(INADDR_ANY);
			addr.sin_port = htons(DSAP_PORTNUM);
			addr.sin_family = AF_INET;

			c[0] = CC_SERVER_SIGNAL;
			LONG2BYTES(thisrec->multiaddr, &c[1]);
			if (sendto(sock, &c[0], 5, 0, (struct sockaddr *)&addr,sizeof(addr))<0)
				err_sys("Can't send UDP packet");
		}
	}

	
	sem_signal(semaphid);                   /* unlock shared memory */
     
	/* Detach shared memory segment */
	if(shmdt((void *)firstrec) < 0)
		err_sys("ERROR: cannot detach shared memory");
}






/* Read file position and current frame number from the shared memory segment */
int read_from_shm(unsigned long id, unsigned long *fpos, int *fnum) 
{
	LOGINTABLE *firstrec, *thisrec;
	int retval = 1;

	if((firstrec = (LOGINTABLE *) shmat(shmid, (char *) 0, 0)) 
		 == (LOGINTABLE *) -1)
		err_sys("ERROR: cannot attach shared memory");

	sem_wait(semaphid);                     /* lock shared memory */
	thisrec = find_pid(firstrec,id);
	if (thisrec == NULL || !(thisrec->status & S_ACTIVE))
	{
		retval = 0;
	}
	else
	{
		retval = 1;
		*fpos = thisrec->filepos;
		*fnum = thisrec->currframe;
	}
	sem_signal(semaphid);

	if(shmdt((void *)firstrec) < 0)
		err_sys("ERROR: cannot detach shared memory");

	return retval;
}


/* Write file position and current frame number into the shared memory segment */
void write_to_shm(unsigned long id, unsigned long fpos, int fnum) 
{
	int retval;
	LOGINTABLE *firstrec, *thisrec;

	if((firstrec = (LOGINTABLE *) shmat(shmid, (char *) 0, 0)) 
		 == (LOGINTABLE *) -1)
		err_sys("ERROR: cannot attach shared memory");

	sem_wait(semaphid);                     /* lock shared memory */
	thisrec = find_pid(firstrec,id);

	if (thisrec==NULL || !(thisrec->status & S_ACTIVE))
	{
		retval = 0;
	}
	else
	{
		retval = 1;
		thisrec->filepos = fpos;
		thisrec->currframe = fnum;
	}
	sem_signal(semaphid);

	if(shmdt((void *)firstrec) < 0)
		err_sys("ERROR: cannot detach shared memory");
}





void read_in_vec() 
{
	int i, limit, currframe;
	unsigned long sz, file_pos;
	struct iovec iov[15];
	struct sess_info *ptr = first;

	err_msg("%d: In read_in_vec()", getpid());

	if (!ptr) sleep_us(500000);

	while(ptr != NULL) {

		if ((ptr->status & S_ACTIVE) && (read_from_shm(ptr->id, &file_pos, &currframe))) {

			ptr->currframe = currframe;
			
	    if (lseek(ptr->fdframe, 4 + 4*currframe, SEEK_SET) < 0)
				err_sys("lseek failed");

	    if (ptr->num_of_frames <= currframe + 15) {  /* End condition */

				err_msg("%d: End condition", getpid()); 


				for (i = currframe; i < ptr->num_of_frames; i++) {
					frameOff[i] = READ_LONG(ptr->fdframe);
				}

				file_pos = frameOff[currframe]; /* Added on 28th May */
				
				sz = ptr->file_len - frameOff[currframe];

				for (i=0; i<(ptr->num_of_frames - currframe - 1); i++) {
					iov[i].iov_base   = ptr->frbuf[i];
					iov[i].iov_len    = frameOff[currframe+i+1] - frameOff[currframe+i]; 
				}

				iov[ptr->num_of_frames - currframe - 1].iov_base = ptr->frbuf[ptr->num_of_frames - currframe - 1];
				iov[ptr->num_of_frames - currframe - 1].iov_len = ptr->file_len - frameOff[ptr->num_of_frames - 1];

				ptr->currframe = ptr->num_of_frames;

				if((lseek(ptr->fd, file_pos, SEEK_SET)) < 0)
					err_sys("ERROR: lseek failed");

				if (readv (ptr->fd, &iov[0], ptr->num_of_frames - currframe) != sz)
					err_sys("readv failed");

				
				err_msg("%d: Session %d ends ....", getpid(), ptr->id);
				server_cleanup(ptr->id);
	    }

	    else {
			 
				/* Normal Condition */
				for (i = currframe; i <= currframe+15; i++) {
					frameOff[i] = READ_LONG(ptr->fdframe);
				}

				file_pos = frameOff[currframe]; /* Added on 28th May */

				sz = frameOff[currframe+15] - frameOff[currframe];

				for (i=0; i<15; i++) {
					iov[i].iov_base   = ptr->frbuf[i];
					iov[i].iov_len    = frameOff[currframe+i+1] - frameOff[currframe+i]; 
				}

				currframe += 15;
				ptr->currframe = currframe;
				
				if((lseek(ptr->fd, file_pos, SEEK_SET)) < 0)
					err_sys("ERROR: lseek failed");

				if (readv (ptr->fd, &iov[0], 15) != sz)
					err_sys("readv failed");

				/* Write the current file position and frame number */
				write_to_shm(ptr->id, frameOff[currframe], currframe);
	    }
		}
		
		ptr = ptr->next;
	}

	

}


inline int make_int(BYTE p1, BYTE p2, BYTE p3, BYTE p4) {

  return ((((int)p1) << 24) + (((int)p2) << 16) + (((int)p3) << 8) + ((int)p4));
	
}


void *get_entity(unsigned char *buff, entity *res) {
	
  res->curr_type = *buff;
  res->frame_num = *(buff+1); 
  res->frame_part = *(buff+2);
  res->num_of_parts = *(buff+3);
  res->size = make_int(*(buff+4), *(buff+5), *(buff+6), *(buff+7));
  res->skip_15_offset = make_int(*(buff+8), *(buff+9), *(buff+10), *(buff+11));

	/*  printf("%c", res->curr_type);*/
  fflush(stdout);
}


void write_to_nw() {
	register int i, j, offset, flag = 0;
	struct sess_info *ptr;
	entity ent, ent0;
	
	/* Timing variables */
	struct timeval time_start, time_stop; /* for elapsed write time */
	double start, stop, deftime, now;
	double global_start, global_stop;
	double nusecs;

	/* code for this function */

	err_msg("%d: In write_to_nw()", getpid());
	
	largebuffer = (BYTE *) malloc(sess_no * GOP_SIZE_LIMIT * sizeof(BYTE));
	
	if (!largebuffer)
		err_sys("malloc() failed");

	pkt_ptr = largebuffer;

	/* Generate the timing info */
	deftime = ((double)1000000.0/(double)FRAMES_PER_SEC);

	/* get the current time */
	timerclear(&time_start);

	if (gettimeofday(&time_start, (struct timezone *) 0) < 0)
		err_sys("ERROR: gettimeofday() failed");

	start = ((double) time_start.tv_sec) * 1000000.0 + (double)time_start.tv_usec;


	for (i=0; i<15; i++) {
		ptr = first;

		if (!ptr) sleep_us(500000);  /* For reducing the server overload */
		
		while(ptr != NULL) {
	    if (ptr->currframe == ptr->num_of_frames)
				blocknum = (ptr->currframe / 15) + 1;
	    else
				blocknum = (ptr->currframe / 15);

	    if (ptr->currframe == ptr->num_of_frames) {
				if (i == (ptr->num_of_frames - ((ptr->currframe/15) * 15)))
					ptr->on = 0;
	    }
			
	    if (ptr->status & S_ACTIVE && ptr->on) {
				
				/* set this flag if we are writing frames for any sessions */
				flag = 1;
				offset = 0;
				
				get_entity(ptr->frbuf[i], &ent0);
				ent = ent0;
				
				for (j=0; j<ent0.num_of_parts; j++) {

					offset += sizeof(entity);
					write_entity(ptr->frbuf[i]+offset,
											 ent.size,
											 (struct sockaddr *) &(ptr->to),
											 addrsize,
											 &ent,
											 blocknum,
											 ptr->id);

					offset += ent.size;

					if (ent.curr_type == 'A')
						j --;
					
					if (j == ent0.num_of_parts-1) {
						break;
					}
					else
						get_entity(ptr->frbuf[i] + offset, &ent);

				}
				
	    }
	    ptr = ptr->next;
		}
		
		/* get time */
		timerclear(&time_stop);
                
		if (gettimeofday(&time_stop, (struct timezone *) 0) < 0)
	    err_sys("ERROR: gettimeofday() failed");
                
		stop = ((double) time_stop.tv_sec) * 1000000.0 + (double)time_stop.tv_usec; 
             
		if ( (nusecs =( (deftime*i) - (stop-start) )) >= 0) {
	    sleep_us(nusecs);
		} 
		else  {
	    /* err_msg("Warning: Server overload at frame level. This might not be critical."); */
		}

		if (!flag) {
	    break;
		}
       
		flag = 0; 

	}
		
	free (largebuffer);
}





/* Reads the whole header and initializes the entities */
#ifdef COMMENT

void read_first_hdr(int fd, ent *e) { 

    e->curr_type = READ_UCHAR(fd);
    e->frame_num = READ_UCHAR(fd); 
    e->frame_part = READ_UCHAR(fd); 
    e->num_of_parts = READ_UCHAR(fd); 
    e->size = READ_INT(fd); 
    e->nextfif = READ_INT(fd); 
}



/*
 * Reads data for all sessions in the linked list.
 */
void read_data(list_pos entry[]) {
	struct sess_info *ptr = first;
	size_t nbyte;
	list_pos list[5];
	int i = 0, j;
	unsigned long file_pos;
	
	while(ptr != NULL) {

		if (ptr->status & S_ACTIVE) {
			ptr->no_ent = 0;
			ptr->count = 1;
			
			/* Read current file position from shared memory */
			file_pos = filepos(ptr->id,entry);
			
			if((lseek(ptr->fd, file_pos, SEEK_SET)) < 0)
				err_sys("ERROR: lseek failed"); 
			
			ptr->no_of_frames = READ_UCHAR(ptr->fd);
			
			read_first_hdr(ptr->fd, ptr->hdr);
			
			for (i=0; i<NSERVERS; i++) {
				if (entry[i].id == ptr->id) {
					entry[i].id = ptr->id;
					entry[i].file_pos = ptr->hdr->ent[0].next_header_offset;
					break;
				}
			}
		
			/* Write position of file pointer to shared memory for next process */
			/*filepos(ptr->id, ptr->hdr->ent[0].next_header_offset,'w');*/
       
			nbyte = (ptr->hdr->ent[0].next_header_offset) -
				(ptr->hdr->ent[0].offset);

			if((j=read(ptr->fd, ptr->buf, nbyte)) != nbyte) {
				fprintf(stderr, "ptr id = %d status= %d\n", ptr->id, ptr->status); 
				err_sys("ERROR: read from file failed, read %d bytes, nbyte = %d", j, nbyte);
			}
			/* If we are at the end of a file, make request to do cleanup */
			if(ptr->hdr->ent[0].next_header_offset == ptr->file_len)
			{
				err_msg("Requested an EOF at %d\n", ptr->id);
				server_cleanup(ptr->id);
			}
		}
		ptr = ptr->next;
	}
}

#endif


#ifdef COMMENT
/*
 * Sends entities in correct order to the corresponding multicast group.
 * This is where timing information is going to need to be inserted.
 */
void write_data(list_gop gop[]) {
	int i, j, n = 0, curr_ent;
	struct timespec ntv;
	unsigned char last_type;
	int flag = 0;
	int count = 1;
	size_t nbyte;
	long offset = 0;
	struct sess_info *ptr = first;
	struct timeval time_start, time_stop; /* for elapsed write time */
	double start, stop, deftime, now;
	double global_start, global_stop;
	double nusecs;
	short gop_num;
	

	largebuffer = (unsigned char *) malloc(sess_no * GOP_SIZE_LIMIT * sizeof(unsigned char));
	pkt_ptr = largebuffer;

	/* Generate the timing info */
	deftime = ((double)1000000.0/(double)FRAMES_PER_SEC);

	/* get the current time */
	timerclear(&time_start);
	if (gettimeofday(&time_start, (struct timezone *) 0) < 0)
		err_sys("ERROR: gettimeofday() failed");
	start = ((double) time_start.tv_sec) * 1000000.0 + (double)time_start.tv_usec;

	for (i=0; i < MAX_FRAMES; i++) {

		ptr = first;
		while(ptr != NULL) {
			if ((ptr->status & S_ACTIVE) && ptr->count <= ptr->no_of_frames) {
	   
				/* set this flag if we are writing frames for any sessions */
				flag = 1;
				
				/* Read the GOP number from shared memory */
				sem_wait(semaphid);
				gop_num = gopnum(ptr->id, gop);
				sem_signal(semaphid);
				
				/* the current entity we are dealing with */
				curr_ent = ptr->no_ent;

				/* Write an entire frame (i.e., | I11 | I12 | etc...) */
				for (j=0; j < ptr->hdr->ent[curr_ent].num_of_parts; j++) {
	     
					offset = (ptr->hdr->ent[ptr->no_ent].offset)
						-(ptr->hdr->ent[0].offset);
	
					if ( ptr->count == ptr->no_of_frames) {

						nbyte = (ptr->hdr->ent[(ptr->count)-1].next_header_offset)
							-(ptr->hdr->ent[(ptr->count)-1].offset);

						write_entity(ptr->buf + offset, 
							    nbyte, 
							    (struct sockaddr *) &(ptr->to),
							    addrsize, 
							    &(ptr->hdr->ent[(ptr->count)-1]), 
							    gop_num, 
							    ptr->id);
						
						++(ptr->count);
	     
  
						/* Increment the GOP number and write to shared memory 
							 for next process */
						fprintf (stderr,"\t\tThe current GOP number is %d\n", gop_num);
						fflush(stderr);
						
						sem_wait(semaphid);

						for (n=0; n<NSERVERS; n++) {
							if (gop[n].id == ptr->id) {
								gop[n].gop_num = ++gop_num;
								break;
							}
						}	
						
						sem_signal(semaphid);
					} 

					else {

						nbyte = (ptr->hdr->ent[ptr->no_ent+1].offset)
							-(ptr->hdr->ent[ptr->no_ent].offset);

						write_entity(ptr->buf + offset, 
							    nbyte, 
							    (struct sockaddr *) &(ptr->to),
							    addrsize, 
							    &(ptr->hdr->ent[ptr->no_ent]), 
							    gop_num, 
							    ptr->id);
						
						++(ptr->count);
					}


					/* we want to club the A frames with the frame if they are 
						 interleaved  */


					if ( (ptr->hdr->ent[ptr->no_ent].curr_type == 'A') && 
							 (ptr->hdr->ent[curr_ent].curr_type != 'A') )
						j -= 1;

					/* we will use the next entity next time through the loop */
					++(ptr->no_ent);
				}
			}
			ptr = ptr->next;
			/*err_msg("****************SWITCHING SESSIONS********************"); */
		}
       
		/* get time */
		timerclear(&time_stop);
		if (gettimeofday(&time_stop, (struct timezone *) 0) < 0)
			err_sys("ERROR: gettimeofday() failed");
		stop = ((double) time_stop.tv_sec) * 1000000.0 + (double)time_stop.tv_usec;              
		if ( (nusecs =( (deftime*i) - (stop-start) )) >= 0) {
			sleep_us(nusecs); 
		} 
		else  {
			/*			err_msg("Warning: Server overload at frame level. This might not be critical."); */
		}

		/* if we haven't written any frames, we can break out of the loop */
		if (!flag) {
			break;
		}
       
		flag = 0; 
	}

	free (largebuffer);
}

#endif


/*
 * Writes one entity to the client.
 */
void write_entity(char *buf, 
									int len, 
									struct sockaddr *to, 
									int tolen, 
									entity *ent, 
									short blknum, 
									unsigned long sess_id) { 

  char msg[100];

  if(put_entity(sockfd, buf, len, to, tolen, ent, blknum, sess_id) < 0) 
    err_sys("ERROR: put_entity failed");

  sprintf(msg, "%lu PUT %d", sess_id, blknum);
  err_msg(msg);
  
  return;
}

/*
 * Fragment entity and send to client.
 */
int put_entity(int sock, 
							 char *buf, 
							 int size, 
							 struct sockaddr *to, 
							 int tolen, 
							 entity *ent, 
							 short gopnum, 
							 unsigned long sess_id) { 
  int i,n;
  register int nleft;
  unsigned short nbytes;
  unsigned char *pkt, *p2, STS[8], length[2], gop[2];
  unsigned char *p1, frag_num, total_frags;
	unsigned long stamp;
  struct timeval timestamp;												
      
  nleft = size;

  /* build the packet header for the client */

	p1 = buf;
  frag_num = 0;
  total_frags = size/PKT_DATA; 
  if(size % PKT_DATA) total_frags++;


  while(nleft > 0)  {
    nbytes = (nleft<PKT_DATA)?nleft:PKT_DATA;
    mklength(nbytes, length);


    p2 = pkt_ptr;
    frag_num++;

    memcpy(p2, length, 2); p2+=2;
    mklength(gopnum, gop);
    memcpy(p2, gop, 2); p2+=2;
    memcpy(p2++, &(ent->frame_num), 1); 
    memcpy(p2++, &(ent->frame_part), 1);
    memcpy(p2++, &(ent->num_of_parts), 1);
    memcpy(p2++, &frag_num, 1);
    memcpy(p2++, &total_frags, 1);
    memcpy(p2++, &(ent->curr_type), 1);


    /* make a time-stamp for the client */
		if (gettimeofday(&timestamp, (struct timezone *) 0) < 0)
			err_sys("ERROR: gettimeofday() failed");
		stamp = ((unsigned long) timestamp.tv_sec) * 1000000.0 + (unsigned long)timestamp.tv_usec;

		mktstamp(sess_id, stamp, STS);
    memcpy(p2, STS, 8); p2+=8;
    memcpy(p2, p1, nbytes);

		
		if (sendto(sock, pkt_ptr, nbytes + 18, 0, to, tolen) < 0) {
			err_msg("ERROR: Sendto failed here");
			abort();
			return(-1);  
		}
		
		
    pkt_ptr += nbytes + 18;

    p1 +=  nbytes;
    nleft -= PKT_DATA;  
  }
  return(0);
}

/*
 * Suspends process execution for the given number of microsecs.
 * From R.S.'s APUE.
 * Takes number of microsecs.
 */
void sleep_us(double nusecs) {
	int i;
  struct timeval tval;

  tval.tv_sec = (long)nusecs/1000000;
  tval.tv_usec = (long)nusecs % 1000000;
  i=select(0,NULL,NULL,NULL,&tval);
	/* err_msg("USleep got %d, %d and returned %d", tval.tv_sec, tval.tv_usec, i); */
}
