#include "netdefs.h"
#include "path.h"
#include "vcripc.h"
#include "vcrctrl.h"
#include "server.h"


FILE *fperr;
int semaphid, msemid;
int shmid, mgid;
int *mcast, id;
unsigned long MAXID;
char mgroup[20];
unsigned int mport;
unsigned long multicast_grp;
unsigned short port_num;

void issue_request(char *, unsigned long, unsigned short);
unsigned long alloc_mgroup();
unsigned short alloc_mport();
void logmessage(char *, char *, char *);


void main(int argc, char *argv[]) {
  int ret;
  FILE *infofile;
  char s[100], fname[100], fullpath[100], line[200];
  struct stat file;

  if((fperr=fopen(ERRORLOG, "a"))==NULL) {
    perror("Unable to open Error Log File");
  }

  multicast_grp = alloc_mgroup();
  port_num = alloc_mport();

  issue_request(argv[1], multicast_grp, port_num);

  fclose(fperr);
}


unsigned long alloc_mgroup() {

  if((mgid = shmget(SHMKEY2, sizeof(int), PERMS)) < 0)
    {
      logmessage("ISSUEREQ", "ERR", "can't get shared memory for mcast group");
      exit(1);
    }     

  if((msemid = sem_open(SEMKEY2)) < 0)
    {
      logmessage("ISSUEREQ", "ERR", "can't open semaphore for mcast group");
      exit(1);
    }    

  if((mcast = (int *)shmat(mgid, (char *) 0, 0)) == (int *)-1)
    {
      logmessage("ISSUEREQ", "ERR", "Can't attach shared memory for mcast group");
      exit(1);
    }


  sprintf(mgroup, "%d.%d.%d.%d", 227, 0, 0, *mcast);
  printf ("%s\n", mgroup);

  id = *mcast; /* Session id for simplicity */

  /* Create entry in table */
  sem_wait(msemid);     /* Lock shared memory for write */

  (*mcast)++;

  sem_signal(msemid);   /* Unlock shared memory */

  if(shmdt((void *)mcast) < 0)
    {
      logmessage("ISSUEREQ", "ERR", "Can't detach shared mem for mcast group");
      exit(1);
    }


  return ((unsigned long) inet_addr(mgroup));
}

unsigned short alloc_mport() {

  mport = 12345;
  printf ("%d\n", mport);

  return ((unsigned short) htons(mport));
}


void issue_request(char * name, unsigned long multicast_grp,unsigned short portnum)
{
     LOGINTABLE *loginrec, *firstrec;

     logmessage("ISSUEREQ", "INF", "issuing a request");

     if((shmid = shmget(SHMKEY, NSERVERS*sizeof(LOGINTABLE), 
         PERMS)) < 0)
     {
          logmessage("ISSUEREQ", "ERR", "can't get shared memory");
          exit(1);
     }     

     if((semaphid = sem_open(SEMKEY)) < 0)
     {
          logmessage("ISSUEREQ", "ERR", "can't open semaphore");
          exit(1);
     }    

     if((firstrec = (LOGINTABLE *)shmat(shmid, (char *) 0, 0)) 
         == (LOGINTABLE *)-1)
     {
          logmessage("ISSUEREQ", "ERR", "Can't attach shared memory");
          exit(1);
     }
     /* Create entry in table */
     sem_wait(semaphid);     /* Lock shared memory for write */

     if((loginrec = find_empty(firstrec)) == NULL)
     {
          logmessage("ISSUEREQ", "ERR", "No free entries");
          exit(1);
     }
     logmessage("ISSUEREQ", "INF", "Setting as active");

		 loginrec->status = S_ACTIVE | S_NEWREQUEST;
     strcpy(loginrec->filename,name);

     loginrec->multiaddr = multicast_grp;
     loginrec->portnum = portnum;

     loginrec->ID = id;
     /*printf("Maxid = %d\n", loginrec->ID);*/
     
     loginrec->filepos = 0;
     loginrec->currframe = 0;
     loginrec->blocknum = 1;

     logmessage("ISSUEREQ", "INF", "Got user request");

     sem_signal(semaphid);   /* Unlock shared memory */

     if(shmdt((void *)firstrec) < 0)
     {
       logmessage("ISSUEREQ", "ERR", "Can't detach shared mem");
          exit(1);
     }
}



void logmessage(char *c_facility, char *c_type, char *c_msg)
{
     time_t t;

     t = time(&t);
     if(! LOGLEVEL) return;
       fprintf(fperr,"%s%%%s::%s @ %s", c_facility, c_type, c_msg, ctime(&t)); 
     fflush(fperr);
}

