#ifndef __VCRCTRL_H__
#define __VCRCTRL_H__

/* These are the  client-server interactions between the VCR Client and */
/* the VOD server.*/

/* Application Defaults */
#define TS_SIZE    5                    /* time stamp size in unsigned chars */
#define STS_SIZE   8                    /* size of server time stamp for each packet */
#define PKT_DATA   553
#define CTRL_DATA  544                  /* max size of data on ctrl channel */
#define MAX_CTRL_SIZE 553               /* maximum size of control pkt */

typedef unsigned char BYTE;

#ifdef COMMENT
typedef struct
{
      BYTE gop_num;           /* GOP number */
      BYTE frag_num;          /* fragment number of sequence */
      BYTE total_frags;       /* total number of fragments in sequence */
      BYTE curr_type;         /* "I", "P", "B", "A", "O" */
      BYTE frame_num;         /* i.e., 1st,2nd,etc... frame in GOP */
      BYTE frame_part;        /* i.e., 1st,2nd.etc... frame of curr_type */
      BYTE num_of_parts;      /* total number of parts */
      BYTE STS[STS_SIZE];     /* time stamp for each packet */
      BYTE length[2];         /* Number of bytes sent in data packet */
      BYTE data[PKT_DATA];    /* MPEG-1 system stream data */
} PKT;
#endif

/* Control packets going between client and server*/
typedef struct _ctrl
{
      BYTE ctrltype;
      BYTE action;
      unsigned short length;
      BYTE argument[CTRL_DATA];
} CTRLPKT;

typedef struct _ack
{
      BYTE ctrltype;
      unsigned long mgroup;
      unsigned short mport;
} ACKPKT;

/* Control types */
#define VCRCTRL  0        /* VCR control */
#define SESSCTRL 1        /* session control */
#define SWITCH   2        /* Server instructs the client to switch */
#define ACK 	 3	  /* Acknowlegdement of a request */

/* VCR Control actions */
#define PAUSE	0	  /* Toggle between Pause/Resume */
#define PLAY	1
#define STOP	2
#define FFWD	3
#define RWND	4

/* Session Control Actions */
#define NEW_UNICAST   0   /* new unicast session - video request */
#define NEW_MULTICAST 1   /* new multicast session - video request */
#define TERM    2         /* terminate a session */


#endif /* !__VCRCTRL_H__ */







