#include "netdefs.h"
#include "vcrctrl.h"
#include "vcripc.h"
#include "server.h"
#include "path.h"
#include "error.h"

#define COMMENT

extern int errno;
FILE *fperr;                 /* Global error log file */
int semaphid, msemid;
int shmid, mgid;

int *mcast;

void shut_down();

void logmessage(char *c_facility, char *c_type, char *c_msg)
{
     time_t t;

     t = time(&t);
     if(! LOGLEVEL) return;
       fprintf(fperr,"%s%%%s::%s @ %s", c_facility, c_type, c_msg, ctime(&t)); 
     fflush(fperr);
}

void create_ipcs(void)
{
     LOGINTABLE *loginrec;

     /* Create shared memory segment */
     if((shmid = shmget(SHMKEY, NSERVERS*sizeof(LOGINTABLE), 
         PERMS|IPC_CREAT)) < 0)
     {
          logmessage("VCRSERVD", "ERR", "can't get shared memory");
          exit(1);
     }

     /* Create the semaphore. */
     if((semaphid = sem_create(SEMKEY, 1)) < 0)
     {
          logmessage("VCRSERVD", "ERR", "can't create semaphore");
          exit(1);
     }    


     /* Gotta clear the table here */
     if((loginrec = (LOGINTABLE *)shmat(shmid, (char *) 0, 0))
          == (LOGINTABLE *)-1)
     {
          logmessage("VCRSERVD", "ERR", "can't attach shared memory");
          exit(1);
     }

     bzero(loginrec, NSERVERS*sizeof(LOGINTABLE));

     if(shmdt((void *)loginrec) < 0)
     {
          logmessage("VCRSERVD", "ERR", "can't detach shared memory");
          exit(1);
     }

     /* Get Shared memory for the Multicast Group */
     if((mgid = shmget(SHMKEY2, sizeof(int), PERMS|IPC_CREAT)) < 0)
     {
          logmessage("VCRSERVD", "ERR", "can't get shared memory for mcast group");
          exit(1);
     }

     /* Get semaphore for MGroup */
     if((msemid = sem_create(SEMKEY2, 1)) < 0)
     {
       logmessage("VCRSERVD", "ERR", "can't create semaphore for mcast group");
       exit(1);
     }    

     if ((mcast = (int *) shmat(mgid, (char *) 0, 0)) == (int *)-1) {
       logmessage("VCRSERVD", "ERR", "can't attach shared memory for mcast group");
       exit(1);
     }

     *mcast = 1;
     
     if (shmdt((void *)mcast) < 0) {
       logmessage("VCRSERVD", "ERR", "can't detach shared memory for mcast group");
       exit(1);
     }
}

void remove_ipcs(void)
{
     if(shmctl(shmid, IPC_RMID, (struct shmid_ds *) 0) < 0)
       logmessage("VCRSERVD", "ERR", "can't remove shared memory");

     logmessage("VCRSERVD", "INF", "removed shared memory");

     sem_rm(semaphid);

     if(shmctl(mgid, IPC_RMID, (struct shmid_ds *) 0) < 0)
       logmessage("VCRSERVD", "ERR", "can't remove shared memory for mcast group");

     logmessage("VCRSERVD", "INF", "removed shared memory for mcast group");

     sem_rm(msemid);

}

void shut_down()
{
     logmessage("VCRSERV", "INF", "trying to shutdown");
     remove_ipcs();
     logmessage("VCRSERVD", "INF", "shutting down");
     abort();
}

#ifndef COMMENT
void issue_request(char * name, long multicast_grp, short portnum)
{
     LOGINTABLE *loginrec, *firstrec;

     logmessage("VCRSERVD", "INF", "issuing a request");

     if((firstrec = (LOGINTABLE *)shmat(shmid, (char *) 0, 0)) 
         == (LOGINTABLE *)-1)
     {
          logmessage("VCRSERVD", "ERR", "Can't attach shared memory");
          exit(1);
     }
     /* Create entry in table */
     sem_wait(semaphid);     /* Lock shared memory for write */

     if((loginrec = find_empty(firstrec)) == NULL)
     {
          logmessage("VCRSERVD", "ERR", "No free entries");
          exit(1);
     }
     logmessage("VCRSERVD", "INF", "Setting as active");

     loginrec->new1 = loginrec->new2 = loginrec->active = 1;
     loginrec->req1 = loginrec->req2 = 0;
     loginrec->ctrltype = SESSCTRL;
     loginrec->action = NEW_MULTICAST;
     strcpy(loginrec->filename,name);

     loginrec->multiaddr = multicast_grp;
     loginrec->portnum = portnum;

     loginrec->ID = MAXID++;

     loginrec->gopnum = 0;
     loginrec->filepos = 0;
     loginrec->flag = 1;

     logmessage("VCRSERVD", "INF", "Got TVGuide request");

     sem_signal(semaphid);   /* Unlock shared memory */

     if(shmdt((void *)firstrec) < 0)
     {
          logmessage("VCRSERVD", "ERR", "Can't detach shared mem");
          exit(1);
     }
}
#endif

void daemon_start(void)
{
     register int i;
     register int childpid;

     if((fperr=fopen(ERRORLOG, "a"))==NULL)
     {
          fprintf(stderr,"can't open error log file \n");
          fflush(stderr);
          exit(1);
     }

     logmessage("VCRSERVD", "INF", "Attempting to start daemon");

     /* Ignore all signals */
     for(i=0; i<33; i++) {
       if( (i != SIGTERM))
	 signal(i, SIG_IGN);
     }
     errno=0;

     if((childpid=fork()) < 0)
     {
          logmessage("VCRSERVD", "ERR", "Can't fork first child for daemon");
          exit(1);                          /* if not started in background */
                                            /* fork and let parent exit    */
     }
      else if(childpid>0) 
         exit(0);           /* parent */

     /* First child process */
     if(setsid() == -1) 
     {
          logmessage("VCRSERVD", "ERR", "Can't change process group");
          exit(1);
     }

     signal(SIGHUP, SIG_IGN);

     logmessage("VCRSERVD", "INF", "Before second fork");
     if((childpid=fork())<0)
     {
          logmessage("VCRSERVD", "ERR", "Can't fork second child for daemon");
          exit(1);
     }
     else if(childpid > 0) 
         exit(0);            /* first child */

     /* Second child process */
     logmessage("VCRSERVD", "INF", "Daemon start successful");

     /* Close all files */
     for(i=0;i<NOFILE;i++) close(i);
     errno=0;

     chdir("/");
     umask(0);
}


void accept_requests(int sockfd, struct sockaddr *cli, int size) {

	int n;
	unsigned long mcastgrp;
	unsigned short mcastport;
	char clipname[256];
	
	CTRLPKT ctrlpkt;
	ACKPKT  ackpkt;
	
	for (; ;) {

		n = recvfrom (sockfd, (char *) &ctrlpkt, MAX_CTRL_SIZE, 0, cli, &size);
		if (n < 0)
			err_msg("VCR: recvfrom error");

		if (ctrlpkt.ctrltype == SESSCTRL) { /* Session Control */
			switch (ctrlpkt.action) {
			case NEW_MULTICAST: {
				/* Read info and issue start request */
				memcpy(clipname, ctrlpkt.argument, ctrlpkt.length);
				mcastgrp = alloc_mgroup();
				mcastport = alloc_mport();
				issue_request(clipname, mcastgrp, mcastport);

				/* Send ack with group and port info */
				ackpkt.ctrltype = ACK;
				ackpkt.mgroup = mcastgrp;
				ackpkt.mport = mcastport;
			}

			case NEW_UNICAST: {
				
			}

			case TERM: {

			}

			}

			if (sendto(sockfd, (char *) &ackpkt, sizeof(ackpkt), 0, cli, size) != sizeof(ackpkt))
				err_msg("VCR: sendto error");
		}
			
	}
}

main(int argc, char *argv[]) {
	FILE *fp;
	char *pname = argv[0];
	int c, debug = 0;

	/* Control Socket information */
	int ctrl_sock;
	struct sockaddr_in      serv, client;


	
  while((c = getopt(argc, argv, "d")) != EOF) {
    switch(c) {
		case 'd':                                 /* debug */
			debug = 1;
			break;
		case '?':
			err_quit("%s: usage: %s [ -d ]", pname, pname);
    }
  }

  /* If not debug, start as a daemon */
	printf ("DEBUG = %d\n", debug);
  if(debug == 0)  
		daemon_start();  

	create_ipcs();  

	atexit(shut_down);     
	signal(SIGTERM,shut_down);
  
	logmessage("VCRSERVD", "INF", "Starting VCR Server");

	if ((ctrl_sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
		err_msg("VCR: Could not open datagram socket");

	bzero ((char *) &serv, sizeof(serv));
	serv.sin_family = AF_INET;
	serv.sin_addr.s_addr = htonl(INADDR_ANY);
	serv.sin_port = htons(SERV_CTRL_PORT);

	if (bind(ctrl_sock, (struct sockaddr *) &serv, sizeof(serv)) < 0)
		err_msg("VCR: Could not bind to socket");

	accept_requests(ctrl_sock, (struct sockaddr *) &client, sizeof(client));
	
	/* pause();*/
	fclose(fperr);
}






